# import petl
import os
import pytest
from oaa.app_runners.custom_app import CustomAppRunner
from pathlib import Path

# test that a configuration can identify multiple DB sources (queries), pull
# that data into streams (petl.fromdb) and feed into OAA

BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture
def multi_db_stream_config(config):
    config.update(config_file=BASE_DIR / "config-v2-sqlplus.yaml")

    return config


def test_config_app(multi_db_stream_config):
    config = multi_db_stream_config
    assert config.version == "v2"
    assert config.app


@pytest.fixture
def env_set():
    os.environ["LD_LIBRARY_PATH"] = (
        f"{os.environ.get('LD_LIBRARY_PATH', '')}:/usr/lib/oracle/10.2.0.5/client64/lib/"
    )
    os.environ["LD_LIBRARY_PATH"] = (
        f"{os.environ['LD_LIBRARY_PATH']}:/usr/lib/oracle/10.2.0.5/client/lib/"
    )
    os.environ["PATH"] = f"/usr/lib/oracle/10.2.0.5/client64/bin:{os.environ['PATH']}"

    os.environ["DB_USERNAME"] = "system"
    os.environ["DB_PASSWORD"] = "oracle"
    os.environ["DB_SERVER"] = "localhost"
    os.environ["DB_PORT"] = "1521"
    os.environ["DB_SERVICE_NAME"] = "xe.oracle.docker"
    # os.environ['PATH'] = $HOME/bin:$PATH


@pytest.fixture
def runner(env_set, multi_db_stream_config):

    yield CustomAppRunner()


@pytest.mark.skipif(
    not os.environ.get("DB_PASSWORD"),
    reason="requires DB_PASSWORD environment variable",
)
def test_fetch_module_list(runner):
    # from oaa.settings_v2 import VezaAPIConnection
    from oaa.settings import config

    source = config.sources["sqlplus"]
    stream = source.get_stream()

    assert stream
    assert stream.nrows() >= 2
    # assert stream.nrows() == 15



import pytest
from oaa.app_runners.custom_app import CustomAppRunner
from oaa.hooks.decorators import HOOK_EVENTS, OAAHookEvent
import logging
logging.basicConfig(log_level=logging.debug)
logger = logging.getLogger(__name__)


@pytest.fixture
def _config(config):
    HOOK_EVENTS[OAAHookEvent.PRE_TRANSFORM]['results'] = []
    config.update(config_file='./tests/config-with-hooks.yaml')

    yield config
    config.reset()

# If hooks are defined in a logic module, then we should be able to find them
# in oaa.app_runners.decorators.HOOK_EVENTS


def test_pre_transform_in_hooks(_config):
    assert _config.sources
    assert HOOK_EVENTS[OAAHookEvent.PRE_TRANSFORM]
    import myhooks
    assert myhooks.do_something in HOOK_EVENTS[OAAHookEvent.PRE_TRANSFORM]['hooks']
    assert HOOK_EVENTS[OAAHookEvent.PRE_TRANSFORM]['results'] == []


def test_hooks_run(_config):
    assert _config.sources
    runner = CustomAppRunner()
    runner.run(push_to_oaa=False)

    assert 'wacka' in HOOK_EVENTS[OAAHookEvent.PRE_TRANSFORM]['results']



# import petl
import os
import pytest
from oaa.app_runners.custom_app import CustomAppRunner
from pathlib import Path

# test that a configuration can identify multiple DB sources (queries), pull
# that data into streams (petl.fromdb) and feed into OAA

BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture
def multi_db_stream_config(config):
    config.update(
        config_file=BASE_DIR / "config-v2-2.yaml",
        config_format="yaml",
        config_version="v2",
    )

    return config


def test_config_app(multi_db_stream_config):
    config = multi_db_stream_config
    assert config.version == "v2"
    assert config.app


@pytest.fixture
def runner(multi_db_stream_config):

    yield CustomAppRunner()


def test_transformer_api(runner, multi_db_stream_config):
    source = multi_db_stream_config.sources["random_users"]
    runner._run_transform(source)
    assert source.stream

    for r in source.records:
        # print(r.name)
        assert isinstance(r["name"], str)
        assert r["unique_id"]


def test_api_responses(runner):
    """TODO: Docstring for test_api_responses.

    :multi_db_stream_config: TODO
    :returns: TODO

    """

    runner.run(push_to_oaa=False)

    oaa_app = runner._provider_object
    assert oaa_app
    assert len(oaa_app.local_users) == 10
    # there should be more users now


def test_fetch_module(runner):
    from oaa.settings_config import APIConnection

    connection = APIConnection(
        uri="https://randomuser.me/api/",
        method="get",
        params={"page": 1, "results": 8, "seed": "wonky"},
        max_results=15,
        fetch_module="api_fetch",
    )
    results = connection.fetch()

    assert results.nrows() == 15



import pytest
from oaa.settings import config
from oaa.settings_utils import SourceType
# Config file config-v2-csv-push describes pushing a csv file directly
# to Veza via the CSV uploader. It still needs a mapping_filepath because it
# will create and update the mapping on the datasource.


@pytest.fixture
def provider_id():
    yield "01994dbd-5551-729b-ab82-8d5714cee692"


@pytest.fixture
def _config():
    config.update(config_file="./tests/config-v2-csv-push.yaml")

    return config


def test_config_settings(_config, provider_id):
    """TODO: Docstring for test_config_settings.

    :multi_db_stream_config: TODO
    :returns: TODO

    """
    assert _config.sources["users"].source_type == SourceType.csv_push
    assert _config.sources["users"].csv_provider_id == provider_id


def test_get_none_provider(_config):
    config.sources["users"].csv_provider_id = "invalid"
    api_mapping = _config.sources["users"].get_api_mapping()
    assert api_mapping is None


def test_get_mapping_from_api(_config):

    api_mapping = _config.sources["users"].get_api_mapping()

    assert api_mapping.template_type == "hris"
    assert api_mapping.column_mappings

    for cm in api_mapping.column_mappings:
        assert cm.column_name or cm.column_name == ""


def test_invalid_csv_to_mapping(_config, provider_id):
    """TODO: Docstring for test_config_settings.

    :multi_db_stream_config: TODO
    :returns: TODO

    """
    assert _config.sources["users_invalid"].source_type == SourceType.csv_push
    assert _config.sources["users_invalid"].csv_provider_id == provider_id

    # from oaa.app_runners.custom_app import CustomAppRunner
    # runner = CustomAppRunner()

    # runner.validate_source_to_mapping(

    source = _config.sources["users_invalid"]
    is_valid, problems = source.validate_to_api_csv_mapping()

    assert is_valid is False


def test_valid_csv_to_mapping(_config, provider_id):
    """TODO: Docstring for test_config_settings.

    :multi_db_stream_config: TODO
    :returns: TODO

    """
    assert _config.sources["users"].source_type == SourceType.csv_push
    assert _config.sources["users"].csv_provider_id == provider_id

    source = _config.sources["users"]
    is_valid, problems = source.validate_to_api_csv_mapping()

    for prob in problems:
        print(prob)
    assert is_valid is True


def test_run_runner(_config, provider_id):
    assert _config.is_csv_push
    from oaa.app_runners.custom_app import CustomAppRunner
    from oaaclient.templates import CustomApplication

    runner = CustomAppRunner()
    runner.run()

    # assert isinstance(runner._provider_object, CustomApplication)

    for source_name, source in config.sources.items():
        assert source.csv_provider_id
        provider = source.get_provider()
        assert isinstance(provider, dict)
        assert provider["id"] == provider_id
    assert runner._get("_created_oaa_objects") is False

    assert runner._get("_created_provider") is False



import pytest
from oaa.app_runners.custom_app import CustomAppRunner
from oaa.hooks.decorators import HOOK_EVENTS, OAAHookEvent
import logging
logging.basicConfig(log_level=logging.debug)
logger = logging.getLogger(__name__)


@pytest.fixture
def _config(config):
    HOOK_EVENTS[OAAHookEvent.LCM_CREATE_USER]['results'] = []
    config.update(config_file='./tests/config-lcm-sources.yaml')

    yield config
    config.reset()

# If hooks are defined in a logic module, then we should be able to find them
# in oaa.app_runners.decorators.HOOK_EVENTS


def test_lcm_create_user_in_events(_config):
    # runner = CustomAppRunner()
    # runner.run(push_to_oaa=False)
    assert _config.sources
    import lcm_test
    assert HOOK_EVENTS[OAAHookEvent.LCM_CREATE_USER]['hooks']
    assert lcm_test.hook_create_user in HOOK_EVENTS[OAAHookEvent.LCM_CREATE_USER]['hooks']
    assert HOOK_EVENTS[OAAHookEvent.LCM_CREATE_USER]['results'] == []


def test_hooks_run(_config):
    assert _config.sources
    runner = CustomAppRunner()
    runner.lcm_create_user({'username': 'kevin'})
    assert ['kevin'] == HOOK_EVENTS[OAAHookEvent.LCM_CREATE_USER]['results']



from .runner_base import BaseRunner
from oaaclient.templates import CustomIdPProvider
from oaaclient.templates import OAAPropertyType
from oaa import oaa_utils
import sys
import logging
from oaaclient.client import OAAClientError
logger = logging.getLogger(__name__)

#########################################################################
#  The Main CLI object. We put it here so that the entrypoint script is
#  simpler.
#########################################################################


class IdPRunner(BaseRunner):

    runner_type = 'IDP'

    def run(self):
        """Import a custom IdP

        This should work similar to run-hris.

        :returns: TODO

        """
        # ingesting a custom IdP
        # get client
        # create
        # create a CustomIdPProvider to represent your IdP. This can be named
        # generically or specific to the environment if you have

        # multiple namespaces to model. idp_type will typically be the
        # technology/vendor for the provider.
        provider_name = self._provider_name
        idp = CustomIdPProvider("My IdP", domain="example.com",
                                idp_type="custom_idp")

        # Add custom property this is the same as for HRIS
        idp.property_definitions.define_user_property("region",
                                                      OAAPropertyType.STRING)
        idp.property_definitions.define_user_property("is_contractor",
                                                      OAAPropertyType.BOOLEAN)

        # What is the minimum set of attributes for user on a custom IdP?
        idp.add_user("mrichardson",
                     full_name="Michelle Richardson",
                     email="mrichardson@example.com")

        evargas_user = idp.add_user("evargas")
        evargas_user.full_name = "Elizabeth Vargas"
        evargas_user.email = "evargas@example.com"

        # Create Groups
        idp.add_group("developers")
        idp.add_group("sec-ops")
        idp.add_group("everyone", full_name="All Company Employees")

        # different ways to add users to groups and groups to users

        for username in idp.users:
            idp.users[username].add_groups(["everyone"])

        evargas_user.add_groups(["developers", "sec-ops"])
        idp.users["mrichardson"].add_groups(["developers"])

        # Veza CustomIdP supports tracking the AWS Roles a user can assume. For
        # users who can assume roles Veza can calculate
        # their effective permissions to AWS resources based on the role(s)
        # roles are added by ARN
        arn_list = [
            "arn:aws:iam::123456789012:role/role001",
            "arn:aws:iam::123456789012:role/role002"
        ]
        idp.users["mrichardson"].add_assumed_role_arns(arn_list)

        veza_con = oaa_utils.get_oaa_client(runner=self)

        # get or create provider
        provider = veza_con.get_provider(provider_name)
        try:
            data_source_name = f"{idp.name} ({idp.idp_type})"
            response = veza_con.push_application(provider['name'],
                                                 data_source_name=data_source_name, # noqa E501
                                                 application_object=idp,
                                                 save_json=True
                                                 )

            if response.get("warnings", None):
                # Veza may return warnings on a successful uploads. These are
                # informational warnings that did not stop the processing
                # of the OAA data but may be important, for example: AWS role
                # ARNs assigned to users that Veza has not discovered
                logger.warn("-- Push succeeded with warnings:")

                for e in response["warnings"]:
                    logger.warn(f"  - {e}")
        except OAAClientError as e:
            logger.error(f"-- Error: {e.error}: {e.message} ({e.status_code})",
                         file=sys.stderr)

            if hasattr(e, "details"):
                for d in e.details:
                    logger.eror(f"  -- {d}", file=sys.stderr)



"""
Utility for retrieving configuration values from multiple sources.

See Also:
    docs/reference/what-is-the-modules-folder.md
    docs/howtos/how-to-write-a-fetch-function.md
"""

import os
from dotenv import load_dotenv

load_dotenv()


def params_or_env(connection, source, key, default=None):
    """
    Get a configuration value from environment, connection params, or source params.

    Lookup order (first match wins):
        1. Environment variable
        2. connection.params[key]
        3. source.params[key]
        4. default value

    Args:
        connection: Connection config object (may have .params dict)
        source: Source config object (may have .params dict)
        key: The key to look up
        default: Value to return if key not found (default: None)

    Returns:
        The value if found, otherwise the default value

    Example:
        # Get API key with fallback
        api_key = params_or_env(connection, source, 'API_KEY', default='default-key')

        # Get required value (will raise KeyError if not found)
        token = params_or_env(connection, source, 'TOKEN') or error()
    """
    # 1. Check environment first (highest priority)
    if key in os.environ:
        return os.environ[key]

    # 2. Check connection params
    if connection and connection.params and key in connection.params:
        return connection.params[key]

    # 3. Check source params
    if source and source.params and key in source.params:
        return source.params[key]

    # 4. Return default if not found anywhere
    return default



import sys
import requests
import dotenv
from functools import wraps
from datetime import datetime, timedelta
import os
import logging
from .base_url_session import BaseURLSession
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)
logger = logging.getLogger(__name__)


dotenv.load_dotenv()

tokenUrl = ''


def get_or_refresh(func):
    '''This is a decorator that handles the in-memory caching of the token and
    refreshing when necessary.
    '''
    token = None
    expires_at = None

    @wraps(func)
    def wrapper(*args, **kwargs):
        nonlocal token
        nonlocal expires_at
        requests_session = kwargs.get('requests_session_cls', requests.Session())

        if not token or datetime.now() >= expires_at:
            logger.debug('getting a new token')
            token = func(*args, **kwargs)

            if not token:
                logger.error('could not fetch token')

                return
            offset = timedelta(seconds=token['expires_in'])
            # set the expiration time to now + expires_in(seconds) - 60 seconds
            # this gives a little bit of a buffer
            expires_at = datetime.now() + offset - timedelta(seconds=60)

        access_token = token['access_token']
        print('access_token', access_token)
        requests_session.headers.update({'Authorization': f'Bearer {access_token}'})

        return requests_session

    return wrapper


@get_or_refresh
def get_token(client_id=None, client_secret=None,
              audience=None, token_url=None, **kwargs):
    '''This script assumes that the `requests` library is installed and that
    the `OAUTH2_CLIENT_ID` `OAUTH2_CLIENT_SECRET`, `OAUTH2_API_AUDIENCE`, and
    `OAUTH2_TOKEN_URL` values are all defined as environmental variables. They
    can also be passed into this function as keyword values.
    >>> from client_exchange import get_token
    >>> token = get_token()

    '''

    payload = {
        'grant_type': 'client_credentials',
        'client_id': os.getenv('OAUTH2_CLIENT_SECRET', client_id),
        'client_secret': os.getenv('OAUTH2_CLIENT_SECRET', client_secret),
        # 'audience': os.getenv('OAUTH2_API_AUDIENCE', 'audience')
    }

    overridable_options = ['audience', 'scope', 'username', 'session_id']
    overridable_options.extend(payload.keys())

    for opt in filter(lambda o: o in kwargs, overridable_options):
        payload[opt] = kwargs[opt]

    # token_url can be overriden
    token_url = os.getenv('OAUTH2_TOKEN_URL', token_url)
    logger.debug('token_url: %s', token_url)
    res = None
    try:
        res = requests.post(token_url, data=payload)

        if res:
            return res.json()
    except Exception as exc:
        logger.error(exc)

    logger.error(res)


get_token_client_credentials = get_token

@get_or_refresh
def get_token_password(client_id=None, client_secret=None,
                       audience=None, username=None,
                       password=None, token_url=None, **kwargs):
    # creds = (SERVICENOW_CLIENT_ID, SERVICENOW_CLIENT_SECRET)
    payload = {
        'grant_type': 'password',
        'client_id': client_id,
        'client_secret': client_secret,
        'username': username,
        'password': password
    }
    response = requests.post(token_url,
                             data=payload)
                             # data={'grant_type': 'client_credentials'})

    print('response', response.json())

    return response.json()['access_token']




import sys
import json
import logging
import time
from datetime import timedelta
from urllib.parse import urljoin
import petl
from oaa.hooks.decorators import OAAHookEvent, hook
from oaa.oaa_utils import get_oaa_client
from oaaclient.client import OAAResponseError
from faker import Faker
logger = logging.getLogger(__name__)

fake = Faker()


@hook(event=OAAHookEvent.POST_TRANSFORM, target='users')
def hook_name(event, source=None, **kwargs):

    '''
    hook_name is a hook for doing some custom thing at this point in the flow.
    '''
    # for board_rec in sources['boards'].records:

    # import bpdb; bpdb.set_trace()  # noqa: E702

def wait_for_status_to_appear(datasource_id):
    """TODO: Docstring for wait_for_status_to_appear.
    :returns: TODO

    """
    veza_con = get_oaa_client()
    path = f'/api/v1/providers/datasources/{datasource_id}/parse_status'
    status_response = None
    done = False

    while not done:
        try:
            # logger.info('checking status at %s', path)
            status_response = veza_con.api_get(path)
            done = True
        except OAAResponseError as err:
            # logger.debug('datasource [%s] does not have a status yet', datasource_id)
            # logger.debug(err.details)
            time.sleep(1)

    # import tqdm
    # tqdm.tqdm([a for a in waiter()])

    return status_response

@hook(event=OAAHookEvent.POST_PUSH_OAA)
def hook_name(event, datasource_id, config=None, **kwargs):

    '''
    hook_name is a hook for doing some custom thing at this point in the flow.
    '''
    # for board_rec in sources['boards'].records:
    # /api/v1/providers/datasources/{id}/parse_status

    provider_id = ''
    status_response = wait_for_status_to_appear(datasource_id)
    integration_url = urljoin(str(config.VEZA_URL), f'/app/integrations/custom/{provider_id}?tab=data_sources')
    output = {
        'status': status_response,
        'integration_url': integration_url  
    }
    # json.dump(status, sys.stdout, indent=2)
    print(json.dumps(output, indent=2))


def fetch(connection=None, source=None, **kwargs):
    '''
    Fetch the data. This is the function that is called by oaa.

    '''

    gen_data = generate_data(source.params.get('USER_COUNT', 10))
    # res = write_csv(10)

    return petl.fromdicts(gen_data)


# --- Configuration ---

# Define specific departments to match the business context of your sample
departments = [
    "Human Resources", "Business Development", "Training", "Services", 
    "Research and Development", "IT", "Marketing", "Sales", "Executive", 
    "Legal", "Engineering", "Support"
]

def generate_data(ROW_COUNT=100):
    
    # Generate n=ROW_COUNT rows

    for i in range(ROW_COUNT):
        # 1. Generate Names
        first_name = fake.first_name()
        last_name = fake.last_name()
        full_name = f"{first_name} {last_name}"

        # 2. Generate Username 
        # Pattern: first initial + last name + unique index (starting at 10)
        # We strip special chars from last name just in case (e.g. O'Conner -> OConner)
        clean_last = "".join(filter(str.isalpha, last_name))
        username = f"{first_name[0].lower()}{clean_last.lower()}{i + 10}"

        # 3. Generate Email
        domain = fake.random_element(elements=("example.com", "example.io"))
        email = f"{username}@{domain}"

        # 4. Generate Dates
        # Created within the last 3 years
        created_at_dt = fake.date_time_between(start_date='-3y', end_date='-1y')

        # Last login must be after created_at
        # We add a random time delta to the created date
        days_after = fake.random_int(min=1, max=300)
        last_login_dt = created_at_dt + timedelta(days=days_after)

        # Format dates to ISO 8601 (e.g., 2021-04-08T08:58:54Z)
        created_at = created_at_dt.strftime("%Y-%m-%dT%H:%M:%SZ")
        last_login = last_login_dt.strftime("%Y-%m-%dT%H:%M:%SZ")

        # 5. Other fields

        user_dict = {
            "username": username,
            "email": email,
            "full_name": full_name,
            "created_at": created_at,
            "last_login": last_login,
            "is_active": fake.boolean(),
            "department": fake.random_element(elements=departments)
        }
        # rows.append([username, email, full_name, created_at, last_login, is_active, department])
        yield user_dict



import os
import struct
import pyodbc
from azure.identity import DefaultAzureCredential


'''
Requirements:
    - pyodbc
    - azure-identity
Requires several system libraries:
unixodbc:
  mac:
      - brew install unixodbc
  redhat:
      - yum install unixODBC
  windows:
      - windows - maybe install with instructions at
        https://learn.microsoft.com/en-us/sql/connect/odbc/download-odbc-driver-for-sql-server?view=sql-server-ver17
'''


def fetch(connection=None, source=None, **kwargs):
    credential = DefaultAzureCredential(
        exclude_interactive_browser_credential=False
    )

    token_bytes = credential.get_token(
        'https://database.windows.net/.default'
    ).token.encode('UTF-16-LE')

    token_struct = struct.pack(
        f'<I{len(token_bytes)}s', len(token_bytes), token_bytes
    )

# This connection option is defined by microsoft in msodbcsql.h
    SQL_COPT_SS_ACCESS_TOKEN = 1256

    driver = '{ODBC Driver 18 for SQL Server}'
    HOST = connection.params['FABRIC_HOST']
    DATABASE = connection.params['FABRIC_DATABASE']
    PORT = connection.params.get('FABRIC_PORT', 1433)
    ENCRYPT = connection.params.get('FABRIC_ENCRYPT', 'yes')
    # DATABASE = 'DWTest'

    connection_string = (
        f'DRIVER={driver};'
        f'SERVER={HOST};'
        f'PORT={PORT};'
        f'DATABASE={DATABASE};'
        f'ENCRYPT={ENCRYPT};'
        'Trusted_Connection=no;'
        'Connection Timeout=30'
    )

    conn = pyodbc.connect(
        connection_string, attrs_before={SQL_COPT_SS_ACCESS_TOKEN: token_struct}
    )
    cursor = conn.cursor()

    query = source.params['query']
    cursor.execute(query)

    return cursor



import sys
import logging
from functools import wraps
import petl
from enum import StrEnum
from typing import Optional
from collections.abc import Callable
from pydantic import BaseModel, Field
from scim_filter_petl import compile_scim_filter
from logging.config import dictConfig
dictConfig(
    {
        "version": 1,
        "formatters": {
            "default": {
                "format": "[%(asctime)s] %(levelname)s in %(module)s: %(message)s",
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "default",
            }
        },
        "root": {"level": "DEBUG", "handlers": ["console"]},
    }
)
# logging.StreamHandler(sys.stdout)
logger = logging.getLogger(__name__)


class EnumSCIMAction(StrEnum):
    SCHEMA = 'schema' 
    USERS = 'users' 
    GROUPS = 'groups'
    USER_ROLES = 'user_roles'


class SCIMAction(BaseModel):
    getter: Callable


class App(BaseModel):
    actions: Optional[dict[EnumSCIMAction, SCIMAction]] = Field(default_factory=lambda: {})
    app_schema: Callable = Field(default=lambda request: {})


class SCIM_APPS_MODEL(BaseModel):
    apps: Optional[dict[str, App]] = Field(default_factory=lambda: {})

    def get_app_action(self, appname, action):
        if action == 'schema':
            return self.apps[appname].app_schema

        return self.apps[appname].actions[action].getter


# SCIM_APPS = {}
SCIM_APPS = SCIM_APPS_MODEL()

def scim_function(_func=None, app_label=None, action=None, **kwargs):

    def inner(func):
        
        nonlocal app_label

        @wraps(func)
        def wrapper(*args, **kwargs):
            """A wrapper function"""

            # Extend some capabilities of func
            

            return func(*args, **kwargs)

        # do something before the decorator is returned
        # HOOK_EVENTS[event]['hooks'].append(wrapper)
        

        SCIM_APPS.apps.setdefault(app_label, App())
        SCIM_APPS.apps[app_label].actions[action] = SCIMAction(getter=wrapper)

        return wrapper

    if _func is None:
        return inner
    else:
        return inner(_func)


def scim_schema(_func=None, app_label=None, **kwargs):

    def inner(func):
        
        nonlocal app_label

        @wraps(func)
        def wrapper(request, *args, **kwargs):
            """A wrapper function"""

            # Extend some capabilities of func
            print('in wrapper...', request)
            

            return func(*args, **kwargs)

        # do something before the decorator is returned
        # HOOK_EVENTS[event]['hooks'].append(wrapper)

        SCIM_APPS.apps.setdefault(app_label, App())
        SCIM_APPS.apps[app_label].app_schema = wrapper

        return wrapper

    if _func is None:
        return inner
    else:
        return inner(_func)


def scim_template(schemas=['urn:ietf:params:scim:api:messages:2.0:ListResponse'],
                  totalResults=None,
                  startIndex=None,
                  itemsPerPage=None,
                  resources=[]):

    return {
        'schemas': schemas,
        'totalResults': totalResults,
        'startIndex': startIndex,
        'itemsPerPage': itemsPerPage,
        'Resources': resources 
    }


class SCIMBase:

    """Docstring for SCIMBase. """

    def __init__(self, app_name=None, scim_action=None, *args, **kwargs):
        """TODO: to be defined1. """
        self._app_name = app_name
        self._scim_action = scim_action

    def run_custom(self, request, page_params={}, filters=None, **kwargs):
        # this should handle the request, taking into account any flask request
        # query parameters
        # Run should return a list of results
        raise NotImplementedError()

    def convert_page_params(self, request):
        # raise NotImplementedError()

        return {}

    def get_filters(self, request):
        '''Filter the list of results by using the SCIM filter string. This is
        more complex than it sounds.
        '''
        # from scim2_filter_parser.queries import SQLQuery
        
        # some filters will be compatible with the underlying application,
        # others will not.

        return []

    def get_total_results(self, request, resources=None, filters=None, startIndex=None):
        '''This should get the total results that match the resource type and
        accounting for any added filters.

        If this is not defined, then we may consider returning the count of
        current resources in the query.'''
        # raise NotImplementedError()

        return len(resources)

    # def get_start_index(self, request):
    #     # raise NotImplementedError()

    #     return 0

    def get_start_index(self, request):
        '''
        SCIM pagination uses startIndex and count to do pagination. We may
        need to convert this to be compatible with the underlying connector
        API.
        '''

        return self.convert_page_params(request).get('startIndex', 0)

    # def get_items_per_page(self, request):
    #     raise NotImplementedError()

    def get_items_per_page(self, request):
        return self.convert_page_params(request).get('count', 100)

    def get_one(self, request, obj_id):
        raise NotImplementedError()

    def run(self, request):
        filters = self.get_filters(request)
        resources = self.run_custom(request=request,
                                    filters=filters,
                                    page_params=self.convert_page_params(request))

        # import petl
        stream = petl.fromdicts(resources)

        # Get the SCIm filter query. If defined, use it to filter the results
        # here. In the future, it might be good to pass this filter or some
        # conversion of it to the underlying connector. For now, we'll just
        # filter the raw results here.
        filter_query = request.args.get('filter', '')

        if filter_query:
            logger.debug('filter_query: %s', filter_query)
            filter_function_here = compile_scim_filter(filter_query)
            # logger.debug('filter_function_here: %s', filter_function_here)
            stream = stream.select(filter_function_here)
            # logger.debug('stream %s', stream)

        startIndex = self.get_start_index(request)
        resources = list(petl.dicts(stream))

        number_resources = self.get_total_results(request,
                                                  resources,
                                                  startIndex=startIndex)

        res = scim_template(totalResults=number_resources,
                            startIndex=startIndex,
                            itemsPerPage=self.get_items_per_page(request),
                            resources=resources)

        return res



from oaa.hooks.decorators import OAAHookEvent, hook


@hook(event=OAAHookEvent.LCM_CREATE_USER, keep_results=True)
def hook_create_user(event, user=None, action=None, **kwargs):

    '''
    hook_name is a hook for doing some custom thing at this point in the flow.
    '''
    print('--------- running')

    return user['username'] 



import os
import pytest
from oaa.settings import config as _config


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "requires_db: tests that require database connection (DB_PASSWORD)"
    )


def pytest_collection_modifyitems(config, items):
    skip_db = pytest.mark.skip(
        reason="requires database connection (DB_PASSWORD environment variable)"
    )

    db_password = os.environ.get("DB_PASSWORD")

    for item in items:
        if "oracle" in item.nodeid.lower() or "sqlplus" in item.nodeid.lower():
            if not db_password:
                item.add_marker(skip_db)


@pytest.fixture
def config():
    _config.update()
    yield _config
    _config.reset()



from typing import Optional, Any
import pytest
from pydantic import ValidationError
# Packages imports
from pydantic import BaseModel, field_validator, ValidationInfo, ConfigDict


class A(BaseModel):
    model_config = ConfigDict(validate_default=True)
    var1: int
    var2: Optional[str] = None

    @field_validator('var2', mode='after')
    @classmethod
    def check_var2(cls, value: Any, info: ValidationInfo) -> Any:

        if info.data['var1'] == 2:

            if not value:
                raise ValueError('field required when var1==2')

        return value


def test_pydantic_validators_optional_none():
    data = {
        "var1": 1,
        # 'var2': None,
        # "var2": {'var3': 'hello', 'var4': 'world'}
    }
    # a = A.model_validate(data)
    a = A(**data)
    assert a
    assert a.var1 == 1
    assert a.var2 is None


def test_pydantic_validators_optional_2():
    data = {
        "var1": 2,
        # "var2": {'var3': 'hello', 'var4': 'world'}
    }
    with pytest.raises(ValidationError):
        a = A.model_validate(data)

    data = {
        "var1": 2,
        "var2": 'hello'
    }
    a = A.model_validate(data)
    assert a
    assert a.var2 == 'hello'



# import petl
import os
import pytest
from oaa.app_runners.custom_app import CustomAppRunner
from pathlib import Path

# test that a configuration can identify multiple DB sources (queries), pull
# that data into streams (petl.fromdb) and feed into OAA

BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture
def multi_db_stream_config(config):
    config.update(config_file=BASE_DIR / 'config-v2-veza-api.yaml')

    return config


def test_config_app(multi_db_stream_config):
    config = multi_db_stream_config
    # assert config.version == 'v2'
    assert config.app


@pytest.fixture
def runner(multi_db_stream_config):

    yield CustomAppRunner()


def test_fetch_module_list(runner):
    # from oaa.settings_v2 import VezaAPIConnection
    from oaa.settings import config
    source = config.sources['veza_query_list']
    stream = source.get_stream()

    assert stream
    assert stream.nrows() >= 2
    # assert stream.nrows() == 15


def test_fetch_api_dict(runner):
    from oaa.settings import config
    source = config.sources['veza_query_dict']
    assert id(source.stream) == id(source.stream)
    assert len(source.stream.fieldnames()) > 1

    assert source.stream
    assert source.stream.nrows() >= 1



from oaa.hooks.decorators import hook, OAAHookEvent


@hook(event=OAAHookEvent.PRE_CREATE_OAA,
      target='application_as_resource')
def application_app_mfa_type(event,
                             sources=None,
                             source=None,
                             provider=None,
                             config=None):
    """TODO: Docstring for return_wacko.
    :returns: TODO

    read each application
    - read MFA_TYPE

    """
    # import bpdb; bpdb.set_trace()  # noqa: E702

    # app_mfa_stream = sources['application_mfa_type']._stream
    new_fields = [
        'mfa_enabled_for_internal',
        'mfa_enabled_for_external_and_fed'
    ]

    # s = sources['application_as_resource']._stream
    s = source._stream
    old_fieldnames = set(s.fieldnames())

    for f in new_fields:
        # remove these from the stream, if they are there

        if f in s.fieldnames():
            s = s.cutout(f)

    rprefix = 'right_pfx_'
    s = s.join(sources['application_mfa_type']._stream,
               rprefix=rprefix,
               key='APPLICATION_ID')

    for f in new_fields:
        s = s.rename(f'{rprefix}{f}', f)
        old_fieldnames.add(f)

    s = s.cut(*old_fieldnames)

    sources['application_as_resource']._stream = s

    # print(sources['application_as_resource']._stream)
    # import bpdb; bpdb.set_trace()  # noqa: E702

    return True



"""
Lazy dictionary table utilities for memory-efficient PETL processing.

See Also:
    docs/reference/what-is-lazy-dict-table.md
    docs/howtos/how-to-write-a-fetch-function.md
"""

import petl


class LazyDict:
    """Wraps a dict generator to yield PETL-compatible tuples lazily.

    Extracts headers from the first dictionary and converts subsequent
    dictionaries to tuples for memory-efficient PETL processing.
    """

    def __init__(self, dict_generator):
        """Initialize with a dictionary generator.

        Args:
            dict_generator: Iterator/generator yielding dictionaries
        """
        self.dict_generator = dict_generator
        self.header = None

    def __iter__(self):
        """Yield PETL-compatible tuples from dictionaries."""
        for d in self.dict_generator:
            if not self.header:
                self.header = tuple(d.keys())
                yield tuple(self.header)
            yield tuple(d.get(k) for k in self.header)


def lazy_dict_table(dict_generator):
    """Convert a dictionary generator into a memory-efficient PETL table.

    Use this instead of petl.fromdicts() for large datasets to reduce memory
    usage by 85-90% through lazy evaluation.

    Args:
        dict_generator: Iterator/generator yielding dictionaries

    Returns:
        PETL table with lazy evaluation

    Example:
        from oaa.modules.lazy_dicts import lazy_dict_table

        def fetch(connection=None, source=None, **kwargs):
            def data_generator():
                for record in api.fetch_all():
                    yield record

            return lazy_dict_table(data_generator())

    See:
        docs/reference/what-is-lazy-dict-table.md
    """
    lazy_table = LazyDict(dict_generator)
    return petl.util.materialise.cache(petl.wrap(lazy_table))



import os
import pytest
from pathlib import Path

from oaa.run import main

# from oaa.settings_utils import SourceType
from oaa.settings import Config


requires_db = pytest.mark.skipif(
    not os.environ.get("DB_PASSWORD"),
    reason="requires DB_PASSWORD environment variable",
)


def test_this_config(config):
    assert config


def test_run_help(capsys):

    main(cli_args=["custom_app", "--help"])

    captured = capsys.readouterr()

    output_expected = [
        "oaa",
        "--help",
        "COMMAND is one of the following:",
    ]

    for output in output_expected:
        # print(captured.out)
        assert output in captured.out

    # print('-' * 20)


def test_run_with_config(capsys):

    BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

    main(
        cli_args=[
            "custom_app",
            "show_config",
            "--config",
            str(BASE_DIR / "tests" / "config-v2.yaml"),
        ]
    )

    captured = capsys.readouterr()

    output_expected = [
        '"log_level": "INFO"',
    ]

    for output in output_expected:
        # print(captured.out)
        assert output in captured.out


log_levelS = ["DEBUG", "INFO", "TRACE", "ERROR"]


@pytest.mark.parametrize("log_level", log_levelS)
def test_new_singleton_config(log_level, config):
    os.environ["log_level"] = log_level
    assert os.getenv("log_level") == log_level
    config.update()
    assert Config.get("log_level") == log_level
    assert config.get("log_level") == log_level


@pytest.mark.parametrize("log_level", log_levelS)
def test_new_singleton_config_2(log_level, config):
    os.environ["log_level"] = log_level
    assert os.getenv("log_level") == log_level
    config.update()
    assert Config.get("log_level") == log_level
    assert config.get("log_level") == log_level


def test_config_change(config):

    assert config._CONFIG
    # assert config.SOURCE_TYPE == 'NONE'
    # assert config.SOURCE_TYPE == SourceType.none
    # assert not config.DB_PASSWORD
    # assert not config.DB_USER


@requires_db
def test_load_from_config_file(config):
    config.update(config_file="./tests/config-v2.yaml")

    assert config.log_level == "ERROR"

    should_be = Path(os.path.dirname(__file__)) / "data/sample_mapping_query1.csv"  # noqa E501
    assert config.sources
    assert config.sources["users"].mapping_filepath == should_be


@requires_db
def test_mod_folder(config):
    """
    We can write custom functions for manipulating data as needed.

    A mod function is created by decorating a function with the @transformer
    decorator found in oaa.utils

    The decorator takes two parameters:
    :table: the stream of data
    :config: the configuration object

    The mod function must return a streaming object
    """

    config.update(config_file="./tests/config-v2.yaml")
    assert config.config_file
    assert config.MODULE_DIRECTORY
    mod_folder_should_be = Path(os.path.dirname(__file__)) / "modules"

    assert config.MODULE_DIRECTORY == mod_folder_should_be


@requires_db
def test_validate_command_with_valid_config(capsys):
    BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

    main(
        cli_args=[
            "custom_app",
            "validate",
            "--config",
            str(BASE_DIR / "tests" / "config-v2.yaml"),
        ]
    )

    captured = capsys.readouterr()

    assert "STEP 1" in captured.out
    assert "VALIDATION PASSED" in captured.out or "VALIDATION FAILED" in captured.out


@requires_db
def test_validate_command_with_config_file_short_flag(capsys):
    """Test that -c flag works as alias for --config-file."""
    BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

    main(
        cli_args=[
            "custom_app",
            "-c",
            str(BASE_DIR / "tests" / "config-v2.yaml"),
            "validate",
        ]
    )

    captured = capsys.readouterr()

    assert "STEP 1" in captured.out
    assert "VALIDATION PASSED" in captured.out or "VALIDATION FAILED" in captured.out


@requires_db
def test_validate_command_with_config_file_long_flag(capsys):
    """Test that --config-file flag works."""
    BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

    main(
        cli_args=[
            "custom_app",
            "--config-file",
            str(BASE_DIR / "tests" / "config-v2.yaml"),
            "validate",
        ]
    )

    captured = capsys.readouterr()

    assert "STEP 1" in captured.out
    assert "VALIDATION PASSED" in captured.out or "VALIDATION FAILED" in captured.out


@requires_db
def test_validate_with_missing_app_fails(capsys, tmp_path):
    """Test that config without required app fails validation."""
    invalid_config = tmp_path / "invalid_config.yaml"
    invalid_config.write_text("""
version: v2
runner_type: custom_app
sources: {}
""")

    main(cli_args=["custom_app", "validate", "--config", str(invalid_config)])

    captured = capsys.readouterr()

    assert "VALIDATION" in captured.out



import os
import petl
from dotenv import load_dotenv
load_dotenv()


class CLIRunner:
    def __init__(self,
                 # service_name=None,
                 source=None,
                 connection=None,
                 **kwargs):

        self.source = source
        self.connection = connection

    def conf(self, var_name, default=None):
        return self.connection.params.get(var_name,
                                          self.source.params.get(var_name,
                                                                 os.getenv(var_name, default))) # noqa E501

    def run_cmd(self, cmd, headers=[], **kwargs):
        '''Run cmd return stream of data.'''

        # run sqlplus with credentials, pointing at .sql file
        source = petl.PopenSource(cmd)
        # read data
        # delimiter = self.source.params.get('delimiter', ',')

        # table = petl.fromcsv(source, delimiter=delimiter)
        table = petl.fromcsv(source)

        if headers:
            table = table.pushheader(headers)

        for header in headers:
            table = table.convert(header, lambda v: v.strip())

        return table



import petl
import requests
from dotenv import load_dotenv
from oaa.hooks.decorators import hook, OAAHookEvent
from oaa.hooks.decorators import hook, OAAHookEvent

load_dotenv()


@hook(event=OAAHookEvent.PRE_TRANSFORM, keep_results=True)
def do_something(event, source=None, **kwargs):
    """TODO: Docstring for return_wacko.
    :returns: TODO

    """

    return 'wacka'

def some_function_to_fetch_data(connection=None, source=None, **kwargs):
    return something

def fetch(connection=None, source=None, **kwargs):
    '''
    Fetch the data. This is the function that is called by oaa.

    '''

    res = some_function_to_fetch_data(connection=connection, source=source)

    return petl.fromdicts(res)



import os
import logging
from flask import Flask
from flask import request
from mk_users import logger
import custom_scim_obj
from lcm_base import LCM_APPS, EnumLCMAction
from scim_base import (
        SCIM_APPS,
        scim_function,
        EnumSCIMAction,
        # scim_schema
)

app = Flask(__name__)

'''
Ideas:
    - CSV -> SCIM converter
    - API -> SCIM converter
    - DB -> SCIM converter
    - Use a completely different API schema than SCIM because it is limited.
'''


@app.route('/veza/lcm/v2/<string:app_name>/writeback', methods=['POST'])
def veza_writeback(app_name):
    '''

    '''
    # import bpdb; bpdb.set_trace()  # noqa: E702
    actions = LCM_APPS.get_app_actions(app_name, EnumLCMAction.WRITEBACK)
    # actions = LCM_APPS.apps.get(app_name, {}).get('writeback', [])
    app.logger.info('actions %s', actions)

    results = []

    for action in actions:
        print('action', action.action.__name__)
        results.append(action.action(request.json))

    # app.logger.info('results %s', results)

    return results


if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0', port=os.getenv('FLASK_PORT', 8080))



from oaa.run import main
from pathlib import Path
from oaa.settings import Config
from oaa.settings_utils import SourceType
import pytest
import os


requires_db = pytest.mark.skipif(
    not os.environ.get("DB_PASSWORD"),
    reason="requires DB_PASSWORD environment variable",
)


LOG_LEVELS = ["DEBUG", "INFO", "TRACE", "ERROR"]


def test_run_help(capsys):

    main(cli_args=["custom_app", "--help"])

    captured = capsys.readouterr()

    output_expected = [
        "oaa",
        "--help",
        "COMMAND is one of the following:",
    ]

    for output in output_expected:
        # print(captured.out)
        assert output in captured.out

    # print('-' * 20)


@requires_db
def test_run_with_config(capsys):

    from pathlib import Path
    import os

    BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))

    main(
        cli_args=[
            "custom_app",
            "show_config",
            "--config-file",
            str(BASE_DIR / "config-v2.yaml"),
            "--config-file-yaml",
            "true",
        ]
    )

    captured = capsys.readouterr()

    output_expected = [
        '"log_level"',
        "config_file",
    ]

    for output in output_expected:
        # print(captured.out)
        assert output in captured.out

    # print('-' * 20)


@pytest.mark.parametrize("log_level", LOG_LEVELS)
def test_new_singleton_config(log_level, config):
    os.environ["log_level"] = log_level
    assert os.getenv("log_level") == log_level
    config.update()
    assert Config.get("log_level") == log_level
    assert config.get("log_level") == log_level


@pytest.mark.parametrize("log_level", LOG_LEVELS)
def test_new_singleton_config_2(log_level, config):
    os.environ["log_level"] = log_level
    assert os.getenv("log_level") == log_level
    config.update()
    assert Config.get("log_level") == log_level
    assert config.LOG_LEVEL == log_level


@requires_db
def test_load_from_config_file(config):
    config.update(config_file="./tests/config-v2.yaml")

    assert config.LOG_LEVEL == "ERROR"

    should_be = Path(os.path.dirname(__file__)) / "data/resources_mapping.csv"
    assert config.sources["apps"].mapping_filepath == should_be


@requires_db
def test_mod_folder(config):
    """
    We can write custom functions for manipulating data as needed.

    A mod function is created by decorating a function with the @transformer
    decorator found in oaa.utils

    The decorator takes two parameters:
    :table: the stream of data
    :config: the configuration object

    The mod function must return a streaming object
    """
    from oaa.settings_v2 import ConfigV2

    config.update(config_file="./tests/config-v2.yaml", config_format="yaml")
    assert config.version == "v2"
    assert isinstance(config._CONFIG, ConfigV2)
    assert config.MODULE_DIRECTORY
    assert config.module_directory
    mod_folder_should_be = Path(os.path.dirname(__file__)) / "modules"

    assert config.module_directory == mod_folder_should_be


# ============================================================================
# Tests for ConfigV2.post_update() model_validator
# ============================================================================


def _make_post_update_config(sources: dict):
    """Build a minimal ConfigV2 with the given sources dict.

    Uses source_type=VEZA_API which requires no filepath — avoids the
    destination=None bypass and keeps sources realistic.
    """
    from oaa.settings_config import ConfigV2
    return ConfigV2.model_validate({
        'version': 'v2',
        'runner_type': 'HRIS',
        'sources': sources,
    })


def test_post_update_sets_source_names():
    """source_name is populated from the config key for each source."""
    config = _make_post_update_config({
        'employees': {'source_type': 'VEZA_API'},
        'roles': {'source_type': 'VEZA_API'},
    })

    assert config.sources['employees'].source_name == 'employees'
    assert config.sources['roles'].source_name == 'roles'


def test_post_update_sets_is_csv_push_when_present():
    """is_csv_push is True when at least one source has source_type=csv_push."""
    config = _make_post_update_config({
        'employees': {'source_type': 'VEZA_API'},
        'org_chart': {'source_type': 'CSV_PUSH'},
    })

    assert config.is_csv_push is True


def test_post_update_keeps_is_csv_push_false():
    """is_csv_push stays False when no csv_push sources exist."""
    config = _make_post_update_config({
        'employees': {'source_type': 'VEZA_API'},
        'roles': {'source_type': 'VEZA_API'},
    })

    assert config.is_csv_push is False



import pytest
from oaa import utils
from oaa.app_runners.hris_single_source import HRISRunner
import petl
import logging
from oaa.settings import config as _config  # , SourceType
logging.basicConfig(log_level=logging.debug)
logger = logging.getLogger(__name__)


@pytest.fixture
def config():
    _config.update(SOURCE_MAPPING_FILEPATH='./tests/sample_mapping.csv')
    yield _config
    _config.reset()


@pytest.fixture
def record():
    obj = {
        'EMPL_ADDRESS1': 'number1',
        'EMPL_ADDRESS2': 'number2',
    }
    record = petl.Record(obj.values(), list(obj.keys()))
    yield record


def test_join_fields_with_none(record, config):
    """Verify that passing None values to join fields does ....
    :returns: TODO

    """
    options = {
            'options': ['\n', ['EMPL_ADDRESS1', None]]
    }

    func = utils.join_fields(options=options)

    assert func('hello', record) == 'number1'


def test_join_fields(record, config):
    """Verify that passing None values to join fields does ....
    :returns: TODO

    """
    options = {
            'options': ['\n', ['EMPL_ADDRESS1', 'EMPL_ADDRESS2']]
    }

    func = utils.join_fields(options=options)

    assert func('hello', record) == '\n'.join(['number1', 'number2'])


VALUES_TO_HASH = (
    ('EMPL_ADDRESS2', 'e9ee33815966ca513b249a589e6a7c4e8aca72995aaa153b631027f562e2f7ea'),  # noqa E501
    ('', 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'),
    (None, None),
)


@pytest.mark.parametrize('value_to_hash, hashed_value', VALUES_TO_HASH)
def test_func_name(value_to_hash, hashed_value):
    """ func_name
    do something with the value_to_hash, hashed_value parameter
    """

    hash_value_here = utils.make_hash()(value_to_hash)
    # print(hash_value_here)
    assert hash_value_here == hashed_value


DATE_VALUES_TO_CONVERT = (
    ('01-01-2025', '2025-01-01T00:00:00Z'),
    ('01/01/2025', '2025-01-01T00:00:00Z'),
    ('01-01-25', '2025-01-01T00:00:00Z'),
    ('01/01/25', '2025-01-01T00:00:00Z'),
    ('2025-01-25', '2025-01-25T00:00:00Z'),
    ('2025-01-25T15:00', '2025-01-25T15:00:00Z'),
)


@pytest.mark.parametrize('date_string,should_be_output_string', DATE_VALUES_TO_CONVERT)  # noqa E501
def test_date_convert(date_string, should_be_output_string):

    options = {
        'options': None
    }
    fieldname = 'EMPL_ADDRESS1'
    output_string = utils.date_convert(fieldname=fieldname,
                                       runner=HRISRunner(),
                                       options=options)(date_string)

    assert output_string == should_be_output_string


EMAILS_TO_TEST = (
    ('kajigga@gmail.com', 'kajigga@gmail.com'),
    ('kajigga+junk@gmail.com', 'kajigga+junk@gmail.com'),
    ('KAJIGGA@gmail.com', 'kajigga@gmail.com'),
    ('invalid@company-name.inv', 'invalid@company-name.inv'),
)


@pytest.mark.parametrize('email,should_be', EMAILS_TO_TEST)
def test_email_convert(email, should_be):
    assert utils.email_convert()(email) == should_be



import petl
from anytree import AnyNode


def make_node(record,
              nodes,
              id_field=None,
              parent_id_field=None):

    if id_field not in record.flds:
        raise Exception(f'field {id_field} '
                        'is not defined.')

    if parent_id_field not in record.flds:
        raise Exception(f'field {parent_id_field} '
                        f'is not defined')

    _id = record[id_field]

    payload = {
        'row': record,
    }

    parent_id = record[parent_id_field]
    parent_node = None

    if parent_id:
        parent_node = nodes[parent_id]
        payload['parent'] = parent_node

    node = nodes.setdefault(_id, AnyNode(id=_id, **payload))

    if parent_node:
        node.parent = parent_node

    node.row = record

    return node


def make_tree(nodes,
              sub_nodes_key='sub_resources',
              id_field=None):
    '''Take a list of nodes and add all of the children to a key,
    defaulting to 'sub_resources'
    '''

    for key, node in nodes.items():

        row = {'tree_id': node.row[id_field]}

        row[sub_nodes_key] = [subnode.row for subnode in node.children]
        yield row


def build_tree(table, id_field, parent_id_field,
               sub_fieldname):
    '''
    Take the table, create a tree based on the parent_id_field
    '''
    # Using anytree https://github.com/c0fec0de/anytree

    # root = AnyNode(id="root")
    # s0 = AnyNode(id="sub0", parent=root)
    # s0b = AnyNode(id="sub0B", parent=s0, foo=4, bar=109)
    # s0a = AnyNode(id="sub0A", parent=s0)
    # s1 = AnyNode(id="sub1", parent=root)
    # s1a = AnyNode(id="sub1A", parent=s1)
    # s1b = AnyNode(id="sub1B", parent=s1, bar=8)
    # s1c = AnyNode(id="sub1C", parent=s1)
    # s1ca = AnyNode(id="sub1Ca", parent=s1c)

    nodes = {}

    # for row in records(table):

    for row in petl.records(table):
        make_node(row, nodes,
                  id_field=id_field,
                  parent_id_field=parent_id_field)

    tree = make_tree(nodes,
                     sub_nodes_key=sub_fieldname,
                     id_field=id_field)

    tree_dicts = petl.fromdicts(tree)

    table = table.leftjoin(tree_dicts,
                           lkey=id_field,
                           rkey='tree_id')

    table = table.convert(sub_fieldname, lambda v: v or [])

    return table



import sys
import importlib
import os
import json
from datetime import datetime
from pathlib import Path
from shutil import copyfile
# from typing import Literal
import petl
# from oaaclient.templates import HRISProvider  # , IdPProviderType
from oaaclient.utils import encode_icon_file
from oaaclient.templates import OAAPropertyType
from oaaclient.client import OAAClientError, OAAResponseError
from oaa import utils
from oaa import oaa_utils
# from oaa.settings_utils import RunnerType
from oaa.settings_config import DestinationType
from oaa.settings import config  # , DBConfig  # , load_config
from oaa.utils import jinja_env
from oaa.settings_config import App

from oaa.models import FieldMapping
import logging
# Make the help output directly to the screen rather than getting stuck in a
# pager. See https://stackoverflow.com/questions/58299680/help-gets-stuck-in-command-line-for-fire-module
os.environ["PAGER"] = "cat"
# import inspect

LOGIC_MODULE_REQUIRED_ARGS = ['config', 'streams']

try:
    import pymssql
except Exception:
    pymyssql = None
try:
    import oracledb
except Exception:
    oracledb = None
try:
    import psycopg
except Exception:
    psycopg = None

logger = logging.getLogger(__name__)
petl.config.failonerror = True

# -----


class BaseRunner(object):
    ''' We create a base runner class so that we can re-use logic that is
    shared across the different types.
    '''
    source_type = None

    def __init__(self,
                 dry_run=False,
                 do_custom_attributes=True,
                 save_json=False,  # whether to save a local copy of
                                   # the payload on push
                 config_file=None,
                 config_file_yaml=None,
                 config_content=None,
                 config_url=None,
                 priority_push=False,
                 # env_file=None,
                 **kwargs
                 ):
        self.__kwargs = kwargs

        self._config_file = config_file
        self._config_file_yaml = config_file_yaml
        self._config_content = config_content
        self._config_url = config_url
        self._priority_push = priority_push

        if self._config_file:
            config.update(config_file=self._config_file,
                          source_type=self.source_type)

        if self._config_content:
            config.update(config_content=self._config_content,
                          source_type=self.source_type)

        if self._config_url:
            config.update(config_url=self._config_url,
                          source_type=self.source_type)

        if self._config_file_yaml:
            config.update(config_file=self._config_file_yaml,
                          config_format='yaml',
                          source_type=self.source_type)

        self._dry_run = dry_run
        self._do_custom_attributes = do_custom_attributes
        self._save_json = save_json

        if not config:
            logger.error('configuration is not valid')
        # else:

        #     self._provider_name = self.config.HRIS_VENDOR_NAME

        #     self._datasource_name = f"{self.config.HRIS_VENDOR_NAME} Datasource"  # noqa E501

        # This will store a list of all attributes (source->destiantion)
        self._attributes = {}
        self._custom_attrs_map = {}
        # self.hris = None
        # this makes oaa_utils available on the CLI as oaa-utils
        self.oaa = oaa_utils.FUNCS
        self._provider_object = None
        self._storage = {}
        self._datasource_name = None
        self._datasource = None

        self._post_init()

    def _get(self, key, default=None):
        return self._storage.get(key, default)

    def _set(self, key, value):
        self._storage[key] = value

    def _post_init(self, *args, **kwargs):
        pass

    def _hook_pre_oaa(self, *args, **kwargs):
        '''If a runner overrides this function, it can do data manipulation of
        the stream(s) before the OAA objects are created.
        '''

        return

    def _get_logic_modules(self):
        """Get any custom logic modules for this runner.
        :returns: TODO

        """
        logic_modules = config.LOGIC_MODULES or []

        modules = []

        for module in logic_modules:
            imported_module = importlib.import_module(module)

            try:
                # argsinfo = inspect.getfullargspec(imported_module.run)
                # assert all((f in argsinfo.args for f in LOGIC_MODULE_REQUIRED_ARGS))   # noqa E501
                assert utils.validate_logic_module(imported_module)
            except Exception as err:
                # TODO this is not longer a requirement. Let's release this so
                # it doesn't show as a warning
                logger.debug('you may consider defining a function called `run` '
                             f' in the {module} module and it must take the '
                             'following arguments: '
                             f' {LOGIC_MODULE_REQUIRED_ARGS}')

            modules.append(imported_module)

        return modules

    def utils(self):
        '''
        List utility functions.

        These functions are those that are available as transforms in the
        mapping document.
        '''

        return utils.transformers

    def show_config(self):

        return config._CONFIG.model_dump_json(indent=2)

    def _get_provider_type(self):
        '''Implement this methods to return the provider class to use when
        creating the provider here.
        '''
        raise NotImplementedError

    def _custom_provider_actions(self, provider_object=None):
        '''Action to take on the provider after the object is instantiated,
        before it is sent to Veza.
        '''
        raise NotImplementedError

    def _get_or_create_datasource(self, **kwargs):
        if not self._datasource:
            veza_con = oaa_utils.get_oaa_client(runner=self)
            datasource = veza_con.get_data_source(name=self._datasource_name,
                                                  provider_id=self._provider_object.provider['id'])

            if not datasource:
                datasource = veza_con.create_datasource(name=self._datasource_name,
                                                        provider_id=self._provider_object.provider['id'])

            self._datasource = datasource

        return self._datasource

    def _get_or_create_provider(self, **kwargs):

        if not self._provider_object:
            # Configure the Custom IdP

            if config.is_csv_push:
                self._set('_created_provider', False)

                return
            logger.info(f"Creating provider for {self._provider_name}")
            provider_cls = self._get_provider_type()

            if isinstance(kwargs, App):
                kwargs = kwargs.model_dump()

            options = {}

            if kwargs.get('options'):
                options = kwargs.pop('options')
            kwargs = {k: v for k, v in kwargs.items() if v}
            # provisioning = kwargs.pop('provisioning', options.get('provisioning', False))

            self._provider_object = provider_cls(**kwargs)

            self._custom_provider_actions(provider_object=self._provider_object)

            veza_con = oaa_utils.get_oaa_client(runner=self)
            provider = veza_con.get_provider(self._provider_name)
            # create a new provider if the provider doesn't already exist

            # create_provider_options = {}
            # external_lifecycle_management_type = kwargs.get('external_lifecycle_management_type')

            # if external_lifecycle_management_type:
            #     create_provider_options['external_lifecycle_management_type'] = external_lifecycle_management_type
            #     create_provider_options['provisioning'] = True

            # TODO update the create_provider action to take the 
            # external_lifecycle_management_type value. Set provisioning=True if it
            # is set.
            # {
            #     "name": "KhansenHR2-v4",
            #     "data_plane_id": "019afaed-8540-71a7-b159-4b91dd1144b5",
            #     "external_lifecycle_management_type": "SEND_REST_PAYLOAD",
            #     "provisioning": true,
            #     "custom_templates": [
            #         "application"
            #     ]
            # }

            if not provider:
                logger.info(f"Creating new provider {self._provider_name}")
                try:
                    provider = veza_con.create_provider(self._provider_name,
                                                        self._provider_object.TEMPLATE,
                                                        options=options)
                    logger.info(f"Provider created {provider['id']}")
                except OAAResponseError as err:
                    logger.error(str(err.details))
            self._provider_object.provider = provider

            if config.HRIS_ICON_B64_PATH:
                veza_con.update_provider_icon(provider['id'],
                                              encode_icon_file(config.HRIS_ICON_B64_PATH)) # noqa E501

            logger.info('template: %s', self._provider_object.TEMPLATE)

            # if self._provider_object.TEMPLATE in ('hris', 'idp') and provider.get('provisioning') != provisioning:

            # if config.runner_type in (RunnerType.hris, RunnerType.idp):  # and provider.get('provisioning') != provisioning:

            '''
            if config.destination_type == DestinationType.HRIS:  # and provider.get('provisioning') != provisioning:
                logger.info('updating provisioning=%s', str(provisioning))
                veza_con.update_provisioning_status(provider['id'], provisioning)
                # try:
                #     veza_con.update_provisioning_status(provider['id'], provisioning)
                # except Exception as err:
                #     logger.error(err.details)
            '''

            # self._provider_object = provider
            # self._get_or_create_datasource()

        return self._provider_object

    def _get_push_options(self):
        options = {
            'priority_push': self._priority_push
        }

        return options

    def _push_csv_file(self, provider, datasource, source):

        payload = oaa_utils.make_csv_upload_payload(provider['id'],
                                                    datasource['id'],
                                                    source.stream)

        oaaclient = oaa_utils.get_oaa_client()

        path = f'/api/v1/providers/custom/{provider["id"]}/datasources/{datasource["id"]}:push_csv' # noqa E501
        logger.info('path: %s', path)

        try:
            response = oaaclient.api_post(path, # noqa E501
                                          payload)

            return response
        except Exception as exc:
            logger.error(exc.details)

    def _push_to_oaa(self):
        veza_con = oaa_utils.get_oaa_client(runner=self)
        response = None
        datasource_id = None

        if config.is_csv_push:
            for source_name, source in config.sources.items():
                datasource = source.upsert_datasource()
                datasource_id = datasource['id']

                has_problems, errors = source.validate_to_api_csv_mapping()

                if not errors:
                    response = self._push_csv_file(source.get_provider(),
                                                   datasource,
                                                   source)
                else:
                    logger.error('csv file for source[%s:%s] is not valid',
                                 source_name, source.csv_filepath)
        else:
            try:
                # push the provider information to Veza, log and errors or
                # warnings

                if not self._dry_run:
                    logger.info("Sending to Veza")

                    # import bpdb; bpdb.set_trace()  # noqa: E702
                    datasource = self._get_or_create_datasource()
                    datasource_id = datasource['id']
                    response = veza_con.push_application(self._provider_name,
                                                         self._datasource_name,
                                                         self._provider_object,
                                                         save_json=self._save_json, # noqa E501
                                                         options=self._get_push_options()) # noqa E501
                    # pl = self._provider_object.get_payload()
                    # del pl['custom_property_definition']['domain_properties']
                    # response = veza_con.push_metadata(self._provider_name,
                    #                                   self._datasource_name,
                    #                                   pl,
                    #                                   save_json=self._save_json, # noqa E501
                    #                                   options=self._get_push_options()) # noqa E501

                    if response.get("warnings", None):
                        logger.warning("Push succeeded with warnings")

                        for w in response['warnings']:
                            logger.warning(w)
                    logger.info("Success")
                else:
                    logger.info('Would have sent a payload to Veza')
                    if self._save_json:
                        metadata = self._provider_object.get_payload()
                        ts = datetime.now().strftime('%Y%m%d-%H%M%S')
                        out_name = f'{self._datasource_name}-{ts}.json'
                        with open(out_name, 'w') as f:
                            f.write(json.dumps(metadata, indent=2))
                        logger.info(f'Dry run payload saved to {out_name}')
            except OAAResponseError as e:
                logger.error(e.details)  # import bpdb; bpdb.set_trace()  # noqa: E702
            except OAAClientError as e:
                logger.error(f"Error during push {e.message} {e.status_code}")

                if e.details:
                    from pprint import pprint
                    pprint(e.details)
                    # logger.error(e.details)
                logger.error("exiting with error")
                sys.exit(1)

        return datasource_id, response

    def _get_stream(self, query=None, source=None):
        '''
        The default is to load data from a DB.

        Overide this method to make it do something different.
        '''
        try:
            conn = self._get_connection()

            if conn:
                table = petl.fromdb(conn, query)

                return table
        except Exception as exc:
            logger.error(exc)

            return

    # Consider caching this method. It's only used once, for now, but it may be
    # called multiple times in the future
    def field_mappings(self,
                       output_format=None,
                       field_mapping_filepath=None):
        '''
        Field Mappings

        View all of the current field mappings. This output of this function
        defaults to json. If pass --output=table then it will output it in a
        text table. The output is really wide, especially when there are long
        notes or field_options, so I'm not really sure how useful it is.

        ¯\\_(ツ)_/¯

        '''
        mappings = {}

        if not field_mapping_filepath:
            field_mapping_filepath = config.SOURCE_MAPPING_FILEPATH

        fields = petl.fromcsv(config.SOURCE_MAPPING_FILEPATH)

        # Use pydantic to validate field mappings

        for idx, field in enumerate(fields.dicts()):

            ctx = {'row_num': idx+1}
            field = FieldMapping.model_validate(field,
                                                context=ctx).model_dump()

            mappings[field['source_field'] or field['destination_field']] = field  # noqa E501

        if output_format == 'table':

            return fields.look(200)

        elif output_format == 'json':

            return json.dumps(mappings, indent=2)

        return mappings

    def gen_config(self, app_name, num_sources=5):
        template = jinja_env.get_template('sample_config.yaml')

        context = {
            'app_name': app_name,
            'num_sources': num_sources,
            'os': os
        }

        return template.render(**context)

    def gen_oaa(self, app_name, num_sources=5, folder_location=None, **kwargs):
        # template = jinja_env.get_template('sample_config.yaml')

        # folder_location

        project_dir = Path(folder_location or self.__kwargs.get('folder_location', Path(os.getcwd())))

        logger.info('project folder %s', project_dir)
        templates = [
            ('README.md', f'{app_name}/README.md'),
            ('sample_config.yaml', f'{app_name}/config.yaml'),
            ('sample.env', f'{app_name}/sample.env'),
            ('requirements.txt', f'{app_name}/requirements.txt'),
            ('modules/appname.py', f'{app_name}/modules/{app_name}.py'),
            ('modules/appname_hooks.py', f'{app_name}/modules/{app_name}_hooks.py'),
            ('modules/appname_mocks.py', f'{app_name}/modules/{app_name}_mocks.py')
        ]

        context = {
            'app_name': app_name,
            'num_sources': num_sources,
            'os': os
        }
        try:
            # Define whatever we want the oaa project directory structure here

            for t, destination in templates:

                # create the folders, as needed
                destination = project_dir / destination
                Path(os.path.dirname(destination)).mkdir(parents=True, exist_ok=True)
                logger.info('creating: %s', destination)
                with open(destination, 'w+') as f:
                    f.write(jinja_env.get_template(t).render(**context))

        except Exception as e:
            print("Failed to create oaa project")
            print(e)

        return "Done!"

    # def generate_sample_field_mapping(self, output_filepath):
    #     '''
    #     Generate sample field mapping

    #     This command generates a sample field mapping document at the pathname
    #     given at output_filepath.

    #     '''

    #     holding_dir = os.path.dirname(output_filepath)

    #     if not os.path.exists(holding_dir):
    #         os.makedirs(holding_dir)

    #     # from pathlib import Path
    #     THIS_BASE_DIR = Path(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))  # noqa E501

    #     sample_path = THIS_BASE_DIR / 'data' / 'sample_field_mapping.csv'
    #     copyfile(sample_path, output_filepath)

    def _create_oaa_objects(self, table):
        """Create the OAA objects. Every runner needs to define this.

        :table: TODO
        :returns: TODO

        """
        logger.info('Starting submission, provider name '
                    f'{self._provider_object.name}')

        for prop_name, field in self._custom_attrs_map.items():

            # TODO map type to an actual type
            destination = field['destination_field']

            logger.debug('destination: %s', destination)
            logger.debug('field: %s', field)
            logger.debug('creating custom_property %s as type: %s',
                         destination, field['veza_field_type'].upper())
            upper_field_type = field['veza_field_type'].upper()
            try:
                field_type = getattr(OAAPropertyType, upper_field_type)
            except AttributeError as err:

                valid_types = [p.name for p in OAAPropertyType]
                logger.error(err)
                logger.error('Invalid custom property type %s:  '
                             'Must be one of %s',
                             upper_field_type,
                             valid_types)
                exit()
            self._provider_object.property_definitions.define_employee_property(prop_name, # noqa E501
                                                                                field_type) # noqa E501
        self._create_oaa_objects_custom(table)

    def _create_oaa_objects_custom(self, table):
        raise NotImplementedError

    def _get_connection(self):
        # logger.info(self._connection_string())
        # DB_DIALECT = 'SQLSERVER'

        if config.DB_DIALECT == 'SQLSERVER':
            try:
                connection = pymssql.connect(config.DB_SERVER,
                                             config.DB_USER,
                                             config.DB_PASSWORD,
                                             config.DB_NAME)
            except pymssql.exceptions.OperationalError as err:
                logger.error(err)

                return
        elif config.DB_DIALECT == 'ORACLE':
            '''
            Supported ORACLE Versions

            https://python-oracledb.readthedocs.io/en/latest/user_guide/installation.html#supported-oracle-database-versions # noqa E501
            '''
            # cs = "localhost/orclpdb"
            connection = oracledb.connect(user=config.DB_USER,
                                          password=config.DB_PASSWORD,
                                          dsn=config.DB_SERVER)
        elif config.DB_DIALECT == 'POSTGRES':
            connect_obj = dict(
                user=config.DB_USER,
                password=config.DB_PASSWORD,
                host=config.DB_SERVER,
                port=8307,
                dbname=config.DB_NAME)
            connection = psycopg.connect(**connect_obj)

        return connection



from scim_base import (
        SCIMBase,
        SCIM_APPS,
        scim_function,
        EnumSCIMAction,
        scim_schema
)
from lcm_base import lcm_action, EnumLCMAction
from mk_users import USERS, USERS_DB, GROUPS, GROUPS_DB, logger
from flask import current_app


class SCIMUsers(SCIMBase):

    # def get_total_results(self, request, resources=None, startIndex=None):
    #     number_users = len(USERS)

    #     if startIndex < number_users:
    #         return number_users

    #     return 0

    def get_one(self, request, obj_id=None):
        return USERS_DB[obj_id]

    def convert_page_params(self, request):
        start_index = int(request.args.get('startIndex', 0))
        count = min(int(request.args.get('count', 50)), None)

        return {
            'startIndex': start_index,
            'count': count
        }

    def run_custom(self, request, page_params={}, **kwargs):
        # this should handle the request, taking into account any flask request
        # query parameters

        # example of filter query string
        # filter=userName+eq+"b8eb04a2-4cc9-40e7-a01f-d1d47c2a56bd"&startIndex=1

        start_index = page_params.get('startIndex', 0)
        count = page_params['count']

        results = USERS[start_index:start_index+count]
        logger.info('%s', results)

        return results


class SCIMGroups(SCIMBase):

    # def get_total_results(self, request, resources=None, startIndex=None):
    #     number_groups = len(GROUPS)

    #     if startIndex < number_groups:
    #         return number_groups

    #     return 0

    def get_one(self, request, obj_id=None):
        return GROUPS_DB[obj_id]

    def convert_page_params(self, request):
        start_index = int(request.args.get('startIndex', 0))
        count = min(int(request.args.get('count', 50)), 50)

        return {
            'startIndex': start_index,
            'count': count
        }

    def run_custom(self, request, page_params={}, **kwargs):
        # this should handle the request, taking into account any flask request
        # query parameters
        start_index = page_params['startIndex']
        count = page_params['count']

        results = []

        results = GROUPS[start_index:start_index+count]
        # TODO look at paginating the members sub-attribute to see if Veza will
        # try to fetch the paginated results
        logger.info('%s', results)

        return results


@scim_schema(app_label='appname')
def get_scim_schema(*args, **kwargs):
    """TODO: Docstring for get_scim_schema.

    :flask_request: TODO
    :returns: TODO

    """

    return {
      "totalResults": 4,
      "itemsPerPage": 4,
      "startIndex": 1,
      "schemas": [
        "urn:ietf:params:scim:api:messages:2.0:ListResponse"
      ],
      "Resources": [
        {
          "schemas": [
            "urn:ietf:params:scim:schemas:core:2.0:Schema"
          ],
          "id": "urn:ietf:params:scim:schemas:core:2.0:User",
          "meta": {
            "resourceType": "Schema",
            "created": "2001-01-01T00:00:00+00:00",
            "lastModified": "2001-01-01T00:00:00+00:00",
            "version": "W/\"07e9640ac446d118dce2ef36c9e12c22d2d92be6\"",
            "location": "https://api.scim.dev/scim/v2/Schemas/urn:ietf:params:scim:schemas:core:2.0:User"
          },
          "name": "User",
          "attributes": [
            {
              "name": "userName",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": True,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "externalId",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "name",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "formatted",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "familyName",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "givenName",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            },
            {
              "name": "active",
              "type": "boolean",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "password",
              "type": "string",
              "mutability": "readWrite",
              "returned": "never",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "emails",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": True,
              "multiValued": True,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "value",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": True,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "type",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "primary",
                  "type": "boolean",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            },
            {
              "name": "groups",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": True,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "value",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "$ref",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "display",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            },
            {
              "name": "roles",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": True,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "value",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": True,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "display",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "type",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "primary",
                  "type": "boolean",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            }
          ]
        },
        {
          "schemas": [
            "urn:ietf:params:scim:schemas:core:2.0:Schema"
          ],
          "id": "urn:ietf:params:scim:schemas:extension:enterprise:2.0:User",
          "meta": {
            "resourceType": "Schema",
            "created": "2001-01-01T00:00:00+00:00",
            "lastModified": "2001-01-01T00:00:00+00:00",
            "version": "W/\"07e9640ac446d118dce2ef36c9e12c22d2d92be6\"",
            "location": "https://api.scim.dev/scim/v2/Schemas/urn:ietf:params:scim:schemas:extension:enterprise:2.0:User"
          },
          "name": "User",
          "attributes": [
            {
              "name": "employeeNumber",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            }
          ]
        },
        {
          "schemas": [
            "urn:ietf:params:scim:schemas:core:2.0:Schema"
          ],
          "id": "urn:ietf:params:scim:schemas:core:2.0:Group",
          "meta": {
            "resourceType": "Schema",
            "created": "2001-01-01T00:00:00+00:00",
            "lastModified": "2001-01-01T00:00:00+00:00",
            "version": "W/\"07e9640ac446d118dce2ef36c9e12c22d2d92be6\"",
            "location": "https://api.scim.dev/scim/v2/Schemas/urn:ietf:params:scim:schemas:core:2.0:Group"
          },
          "name": "Group",
          "attributes": [
            {
              "name": "displayName",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": True,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "externalId",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "members",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": True,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "value",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": True,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "$ref",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "display",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            }
          ]
        },
        {
          "schemas": [
            "urn:ietf:params:scim:schemas:core:2.0:Schema"
          ],
          "id": "dev:scim:Office",
          "meta": {
            "resourceType": "Schema",
            "created": "2001-01-01T00:00:00+00:00",
            "lastModified": "2001-01-01T00:00:00+00:00",
            "version": "W/\"07e9640ac446d118dce2ef36c9e12c22d2d92be6\"",
            "location": "https://api.scim.dev/scim/v2/Schemas/dev:scim:Office"
          },
          "name": "Office",
          "attributes": [
            {
              "name": "name",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": True,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "displayName",
              "type": "string",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": False,
              "caseExact": False
            },
            {
              "name": "employees",
              "type": "complex",
              "mutability": "readWrite",
              "returned": "default",
              "uniqueness": "server",
              "required": False,
              "multiValued": True,
              "caseExact": False,
              "subAttributes": [
                {
                  "name": "id",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "$ref",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                },
                {
                  "name": "display",
                  "type": "string",
                  "mutability": "readWrite",
                  "returned": "default",
                  "uniqueness": "server",
                  "required": False,
                  "multiValued": False,
                  "caseExact": False
                }
              ]
            }
          ]
        }
      ]
    }


@scim_function(app_label='appname', action=EnumSCIMAction.USERS)
def appname_users():
    return SCIMUsers


# @scim_function(app_label='appname', action=EnumSCIMAction.SCHEMA)
# def appname_schema():
#     return SCIMSchema


@scim_function(app_label='appname', action=EnumSCIMAction.GROUPS)
def appname_groups():
    return SCIMGroups

@lcm_action(app_label='appname', action=EnumLCMAction.WRITEBACK)
def lcm_writeback(payload):
    current_app.logger.info('payload: %s', payload)
    # import bpdb; bpdb.set_trace()  # noqa: E702

    return payload


@lcm_action(app_label='appname', action=EnumLCMAction.WRITEBACK)
def lcm_writeback2(payload):
    current_app.logger.info('payload v2: %s', payload)
    # import bpdb; bpdb.set_trace()  # noqa: E702

    return payload



import urllib.parse
import logging
from functools import wraps
from unittest.mock import MagicMock
from oaa.modules.params_or_env import params_or_env
logger = logging.getLogger(__name__)


MOCK_RESPONSES = {}


class MockResponse:
    def __init__(self, resp, status_code=200):
        """TODO: Docstring for __init__.

        :resp: TODO
        :returns: TODO

        """
        self.response = resp
        self._status_code = status_code

    # @classmethod
    def json(self, *args):
        return self.response

    @property
    def status_code(self):
        # import bpdb; bpdb.set_trace()  # noqa: E702

        return self._status_code 


def mock_response(_func=None, path=None, status_code=200, **kwargs):

    def inner(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            """A wrapper function"""

            # Extend some capabilities of func
            res = MockResponse(func(*args, **kwargs), status_code=status_code)
            logger.debug('returning mocked response for %s -- %s', path, res)

            return res

        # do something before the decorator is returned
        # HOOK_EVENTS[event]['hooks'].append(wrapper)

        if not path:
            raise Exception(f'You must define an overrided path for mocks({path})')
        ret = None

        logger.debug('preparing mock response for %s', path)
        MOCK_RESPONSES[path] = wrapper

        return MOCK_RESPONSES[path]

    if _func is None:
        return inner

    return inner(_func)


class OAAMockException(Exception):
    pass


def call_response(path):
    """TODO: Docstring for call_response.

    :path: TODO
    :returns: TODO

    """

    if not MOCK_RESPONSES.get(path):
        raise OAAMockException(f'no overridden path provided for {path}')

    return MOCK_RESPONSES.get(path)()


def get_sideeffect(path, *args, **kwargs):
    try:
        path = urllib.parse.urlparse(path).path
    except Exception as e:
        print('path', path, e)
    logger.debug('getting response for %s', path)
    res = call_response(path)

    if res is None:
        raise OAAMockException(f'no mocked response for path {path}')

    return res


def maybe_mock_it(mocked_object, method, connection, source):

    if str(params_or_env(connection, source, 'MOCK_RESULTS')).lower() == 'true':
        setattr(mocked_object, method, MagicMock(side_effect=get_sideeffect))



import json
import timeit
import time
import logging
import traceback
import petl
from .runner_base import BaseRunner
from oaa.settings import config
from oaa import utils
from oaa import oaa_utils
from oaa.models import VezaObjectType
from oaa.settings_config import DestinationType

# from oaa.models import DynamicTranformOptions_Enum
from oaa.hooks.decorators import OAAHookEvent, run_hooks
from oaa.hooks.preflight_checks import run_preflights
from oaaclient.templates import (
    CustomApplication,
    CustomIdPProvider,
    HRISProvider,
    # CustomPermission,
    # OAAPermission,
    OAAPropertyType,
    # LocalRole
)

# from oaaclient.templates import HRISProvider  # , IdPProviderType
petl.config.failonerror = True

logger = logging.getLogger(__name__)


def fmapper(source_f):
    """Take source field, then return rec[source_f]"""

    return lambda val, rec: rec[source_f]


def add_missing_fields(stream, should_be_list, current_list):
    # TODO Test this function
    missing_list = set(should_be_list) - set(current_list)

    for destination_field in missing_list:
        stream = stream.addfield(destination_field, "")

    return stream


# def add_field_if_missing(stream, field):
#     def add_p(r):

#         return ""

#     # I need to remove this. It is causing the stream to be read, and parsed,
#     # every time it is called.
#     # TODO make a addfieldifmissing function

#     if field not in stream.fieldnames():
#         logger.debug('adding field %s', field)

#         stream = stream.addfield(field, add_p)
#     else:
#         logger.debug('%s already in fields', field)

#     return stream


class LCMMixin:
    def lcm_create_user(self, user, action=None, **kwargs):
        return run_hooks(
            event=OAAHookEvent.LCM_CREATE_USER,
            config=config,
            sources=config.sources,
            provider=self._provider_object,
            user=user,
            action=action,
        )

    def lcm_modify_user(self, user, action=None, **kwargs):
        return run_hooks(
            event=OAAHookEvent.LCM_MODIFY_USER,
            config=config,
            sources=config.sources,
            provider=self._provider_object,
            user=user,
            action=action,
        )

    def lcm_get_metadata(self, **kwargs):
        return run_hooks(
            event=OAAHookEvent.LCM_MODIFY_USER,
            config=config,
            sources=config.sources,
            provider=self._provider_object,
        )


class CustomAppRunner(BaseRunner, LCMMixin):
    """
    A Runner for doing pushes of Custom app data to Veza

    If you want to omit custom attributes, even if mapped in the file, pass
    in the argument --do-custom-attributes=False.

    Include --dry-run=True to make this do all of the parsing but not send the
    payload to Veza.

    >>> oaa custom_app --config-file ./path/to/config.yaml run


    """

    runner_type = "Custom"

    def _post_init(
        self,
        *args,
        **kwargs,
    ):

        if not config.valid():
            logger.error("configuration is not valid")
            exit(1)
        else:
            if config.app:
                self._provider_name = config.app.name

                self._datasource_name = f"{config.app.name} Datasource"  # noqa E501

        self._set("provider", None)
        self._set("_created_provider", None)
        self._set("_created_oaa_objects", None)
        self._get_logic_modules()

    def preflight(self):
        """
        Run preflight checks.

        This function will run several preflight checks on the configuration.
        It should be a pretty good indication of the readiness of the script to
        run.

        """
        problems = run_preflights(self)
        logger.debug("preflight problems: %s", problems)

    def validate(self, config=None, test_connectivity=False):
        """Validate the configuration file without running a full sync.

        This command validates the config.yaml file by:
        1. Validating the config structure (Pydantic validation)
        2. Checking that custom connection modules can be imported
        3. Running preflight checks
        4. Optionally testing connectivity to data sources

        :param config: Path to the config.yaml file (or use -c/--config)
        :param test_connectivity: If True, test actual connections to dbs/APIs.
                                 Defaults to False for faster validation.
        :returns: Tuple of (is_valid, messages)

        Example:
            oaa custom_app -c ./config.yaml validate

            # With connectivity testing:
            oaa custom_app -c ./config.yaml validate --test_connectivity=True
        """
        from oaa.hooks.preflight_validate import validate_config

        config_file = config or self._config_file_yaml or self._config_file

        is_valid, messages = validate_config(
            config_file, test_connectivity=test_connectivity
        )

        for msg in messages:
            print(msg)

        return is_valid, messages

    def run(self, push_to_oaa=True):
        return_value = {"status": True, "report": []}
        try:
            # process application
            # get or create provider
            self._get_or_create_provider(**config.app.model_dump())
            # assert provider_obj
            self._get_or_create_datasource()

            assert self._datasource

            self._run_transforms()

            self._create_oaa_objects()

            prepush_responses = run_hooks(
                event=OAAHookEvent.PRE_PUSH_OAA,
                config=config,
                sources=config.sources,
                provider=self._provider_object,
            )

            do_push = True

            if any((r is False for r in prepush_responses)):
                do_push = False

            # RECORD transforms
            run_hooks(
                event=OAAHookEvent.RECORD_TRANSFORM,
                config=config,
                sources=config.sources,
                provider=self._provider_object,
            )

            self._hook_post_oaa()

            response = None
            datasource_id = None

            if push_to_oaa and do_push:
                datasource_id, response = self._push_to_oaa()

            # import bpdb; bpdb.set_trace()  # noqa: E702
            run_hooks(
                event=OAAHookEvent.POST_PUSH_OAA,
                response=response,
                config=config,
                provider=self._provider_object,
                datasource_id=datasource_id,
            )

        except Exception as _err:
            logger.error("error running %s", _err)
            logger.error(traceback.format_exc())
            return_value["status"] = "exception"
            return_value["report"].append(_err)

        return return_value

    def _hook_post_oaa(self):

        # Get the logic modules here and run there

        for module in self._get_logic_modules():
            if hasattr(module, "run"):
                module.run(self._provider_object, sources=config.sources, config=config)

    def _run_transforms(self):
        for name, source in config.sources.items():
            self._run_transform(source)
            # runtime = timeit.timeit(lambda: self._run_transform(source), number=1)
            #
            # if runtime >= 2:
            #     logger.warning('running transforms took a long time (%s seconds).'
            #                    'This is something to look into but not uncommon, '
            #                    'especially if the dataset is very large.',
            #                    runtime)
            #
            run_hooks(
                event=OAAHookEvent.POST_TRANSFORM,
                config=config,
                provider=self._provider_object,
                sources=config.sources,
                source=source,
            )

    def _run_transform(self, source):
        """
        Run the Transforms on the source fields. This function also populates
        the source->destination mapping.
        """

        # Run the OAAHookEvent.PRE_TRANSFORM hooks
        logger.debug("run transforms for source %s", source.source_name)
        run_hooks(
            event=OAAHookEvent.PRE_TRANSFORM,
            config=config,
            provider=self._provider_object,
            source=source,
        )

        run_hooks(
            event=OAAHookEvent.RECORD_TRANSFORM_STREAM,
            config=config,
            source=source,
            provider=self._provider_object,
        )

        # if not source.stream:
        #     logger.error('unable to fetch any data')

        fields_to_cutout = set()

        try:
            field_mappings = self.field_mappings(source=source)
        except Exception as e:
            print(source.source_name)
            print(e)
            # import bpdb; bpdb.set_trace()  # noqa: E702
            # exit()

        # TODO look at using a fieldmap here instead of individual actions
        # see https://petl.readthedocs.io/v1.7.3/transform.html#petl.transform.maps.fieldmap
        # It might be more efficient but would require reworking the utility
        # functions

        # How can I know what fields exists in teh data source without reading
        # from it at least once? I need to add fields sometimes so that later
        # they can be mapped to.

        # then = time.time()

        # Let's get the current list of fieldnames. This incurs a cost upfront,
        # but we can re-use this list later, instead of reading the fieldnames
        # over and over.
        # Let's make this a set so it handles uniques all by itself.

        # import ipdb; ipdb.set_trace()  # noqa: E702
        header_list = petl.header(source.stream)
        # if isinstance(source.stream.header, list):
        #     header_list = source.stream.header
        # elif callable(source.stream.header):
        #     header_list = source.stream.header()
        # else:
        #     header_list = source.stream.fieldnames()

        initial_fieldnames_list = set(header_list)

        for fieldname_label, field in field_mappings.items():
            logger.debug("checking %s [%s]", fieldname_label, field)

            if not field.include:
                continue

            fieldname = field.source_field or fieldname_label

            destination = field.destination_field

            deferred_converts = []

            for transform_obj in field.transforms:
                transform = transform_obj.name
                field_options = transform_obj.options
                func_obj = utils.transformers.get(transform, None)

                if func_obj:
                    func = func_obj["func"]
                    # some functions require special treatment. This usually
                    # means passing in to config value(s) to the function then
                    # returning the inner function to the petl convert/addfield
                    # method.

                    mode = func_obj["mode"]

                    # func = func(options=field.field_options)
                    func = func(options=field_options)

                    if mode == 'convert':
                        if fieldname != destination:
                            # Defer to run on destination after source→dest copy
                            # so the source field is not modified in-place and
                            # other mappings sharing the same source field are
                            # not corrupted.
                            deferred_converts.append(func)
                        else:
                            source._stream = source.stream.convert(fieldname, func, pass_row=True)

                    elif mode == "add":
                        source._stream = source.stream.addfield(destination, func)
                        initial_fieldnames_list.add(destination)
                        # The field was added, so we need to reference the new
                        # fieldname here
                        fieldname = destination
                    elif mode == "select":
                        source._stream = source.stream.select(fieldname, func)
                # elif field.transform == 'ignore':
                elif transform == "ignore":
                    fields_to_cutout.add(fieldname)
                elif transform == "distinct":
                    source._stream = source.stream.distinct(fieldname)

            # source.stream = add_field_if_missing(source.stream, destination)

            logger.debug("transforming field %s", destination)

            if field.is_custom:
                source.custom_attrs_map[destination] = field  # fieldname
            else:
                # When source and destination differ, the destination column
                # is populated via fmapper + any deferred transforms, so read
                # from destination at object-creation time instead of the raw
                # source field.
                if fieldname != destination:
                    source.attributes[destination] = destination
                else:
                    source.attributes[destination] = fieldname

            # ----------------
            # do source -> destination mapping

            if fieldname != destination:
                # the source field is different from the destination

                if destination not in initial_fieldnames_list:
                    source._stream = source.stream.addfield(destination, "")

                source._stream = source.stream.convert(
                    destination, fmapper(fieldname), pass_row=True
                )

                for func in deferred_converts:
                    source._stream = source.stream.convert(destination, func, pass_row=True)

                # import bpdb; bpdb.set_trace()  # noqa: E702

            # Normally the actual mapping happens just before the OAA object is
            # pushed. I want to make that happen here instead, as part of the data
            # stream.
            # end source -> destination mapping

        necessary_destinations = [
            f.destination_field for idx, f in field_mappings.items()
        ]
        initial_fieldnames_list.update(necessary_destinations)
        final_fieldnames_list = list(initial_fieldnames_list)

        # add missing fields
        source._stream = add_missing_fields(
            source.stream, necessary_destinations, initial_fieldnames_list
        )

        # cutout fields we don't want anymore. They aren't removed before
        # this moment because they may be needed in join_fields

        filtered_fieldnames = filter(
            lambda f: f in final_fieldnames_list, fields_to_cutout
        )  # noqa E501

        for fieldname in filtered_fieldnames:
            source._stream = source.stream.cutout(fieldname)

        null_to_blank_func = utils.transformers.get("null_to_blank", None)["func"]

        # for f in final_fieldnames_list:
        #     source.stream = source.stream.convert(f, null_to_blank_func(options={}))
        source._stream = source.stream.convertall(null_to_blank_func(options={}))

    def show_field_mapping(self):
        for n, s in config.sources.items():
            print("=" * 25)
            print("source", n)
            print("=" * 25)

            for fm, f in s.field_mappings().items():
                print("mapping name: ", fm)
                print("source_field: ", f.source_field or fm)
                print("destination_field: ", f.destination_field)
                print("transforms: ", [t.name for t in f.transforms])
                print("")
                print("-" * 25)

    # Consider caching this method. It's only used once, for now, but it may be
    # called multiple times in the future
    def field_mappings(self, output_format=None, source=None):
        """
        Field Mappings moved to the settings.Source object.
        """

        if isinstance(source, str):
            source = config.sources[source]
        mappings = source.field_mappings()

        if output_format == "table":
            return mappings

        elif output_format == "json":
            return json.dumps(mappings, indent=2)

        return mappings

    def _get_provider_type(self):

        if config.destination_type == DestinationType.CUSTOM_APP:
            return CustomApplication
        elif config.destination_type == DestinationType.IDP:
            return CustomIdPProvider
        elif config.destination_type == DestinationType.HRIS:
            return HRISProvider

        if not config.destination_type:
            return CustomApplication

    def _custom_provider_actions(self, provider_object=None):
        # Multiple types can be added by running `add_idp_type` multiple times
        # with different IdPProviderType enums
        # provider_object.system.add_idp_type(config.IDP_PROVIDER_TYPE)
        pass

    def _create_oaa_objects(self):

        if config.is_csv_push:
            self._set("_created_oaa_objects", False)

            return

        for source_name, source in config.sources.items():
            responses = run_hooks(
                event=OAAHookEvent.PRE_CREATE_OAA,
                provider=self._provider_object,
                config=config,
                sources=config.sources,
                source=source,
            )

            # set do_oaa_create to False if any responses are False
            do_oaa_create = not any((r is False for r in responses))

            # if any((r is False for r in responses)):
            #     do_oaa_create = False

            # if do_create:
            #     self._create_oaa_objects()

            logger.info(
                f"Loading {source.destination} records for {source.source_name}"
            )

            if source.destination != VezaObjectType.none:
                for prop_name, field in source.custom_attrs_map.items():
                    # Map type to an actual type
                    destination = field.destination_field

                    logger.debug("destination: %s", destination)
                    logger.debug("field: %s", field)
                    logger.debug(
                        "creating custom_property %s as type: %s",
                        destination,
                        field.veza_field_type.upper(),
                    )
                    upper_field_type = field.veza_field_type.upper()
                    try:
                        field_type = getattr(OAAPropertyType, upper_field_type)
                    except AttributeError as err:
                        valid_types = [p.name for p in OAAPropertyType]
                        logger.error(err)
                        logger.error(
                            "Invalid custom property type %s:  Must be one of %s",
                            upper_field_type,
                            valid_types,
                        )
                        exit()
                    definer_map = {
                        VezaObjectType.LocalUser: "define_local_user_property",
                        VezaObjectType.LocalGroup: "define_local_group_property",
                        VezaObjectType.LocalRole: "define_local_role_property",
                        VezaObjectType.Resource: "define_resource_property",
                        VezaObjectType.User: "define_user_property",
                        VezaObjectType.Group: "define_group_property",
                        VezaObjectType.Employee: "define_employee_property",
                    }
                    attr = definer_map.get(source.destination)

                    if attr:
                        definer = getattr(
                            self._provider_object.property_definitions,
                            definer_map.get(source.destination),
                            None,
                        )

                        if definer:
                            if source.destination == VezaObjectType.Resource:
                                definer(
                                    property_type=field_type,
                                    name=prop_name,
                                    resource_type=source.resource_type,
                                )
                            else:
                                definer(property_type=field_type, name=prop_name)
                            # resource_type=source.resource_type)

            if "row" not in petl.header(source.stream):
                source._stream = source._stream.addrownumbers()

            if source.destination != VezaObjectType.none:
                # import bpdb; bpdb.set_trace()  # noqa: E702

                for r in source.records:
                    logger.debug("%s: working on row %s", source.source_name, r.row)

                    # for f_name in r.flds:
                    #     assert not callable(r[f_name])

                    object_with_required_fields = {}
                    required_fields_tmp = []

                    temp_oaa_obj = None

                    # if source.destination != VezaObjectType.LocalUser:
                    #     import bpdb; bpdb.set_trace()  # noqa: E702

                    # if source.destination != VezaObjectType.none:
                    required_fields_tmp = oaa_utils.REQUIRED_FIELDS[
                        source.destination.name
                    ]  # noqa E501

                    for req_field in required_fields_tmp:
                        # req_field_to_map = req_field

                        if isinstance(req_field, tuple):
                            # one of these fields is required, not both
                            found_in_mapping = 0

                            for f in req_field:
                                if (
                                    f in source.attributes
                                    and source.attributes[f] in r.flds
                                ):  # noqa E501
                                    found_in_mapping += 1
                                    req_field = f

                            if found_in_mapping > len(req_field):
                                logger.error(
                                    "only one of these %s can be mapped",  # noqa E501
                                    req_field,
                                )
                                exit()

                            if found_in_mapping == 0:
                                logger.error("one of %s must be mapped", req_field)  # noqa E501
                                exit()

                        else:
                            if not source.attributes:
                                logger.error(
                                    "something is wrong, source.attributes is empty"
                                )  # noqa E501

                            logger.debug(
                                "attributes in mapping: %s", source.attributes.keys()
                            )

                            req_field_dest = source.attributes.get(req_field)

                            if not req_field_dest:
                                logger.error(
                                    "required field %s is not mapped", req_field
                                )
                                exit()

                            if req_field_dest not in r.flds:
                                # print(source.attributes)
                                logger.error(
                                    "%s not in attributes on record: %s",
                                    req_field_dest,
                                    r.flds,
                                )
                                exit()

                                # exit()
                        object_with_required_fields[req_field] = r[
                            source.attributes[req_field]
                        ]  # noqa E501

                    if do_oaa_create:
                        # change this based on the source.destination

                        if source.destination == VezaObjectType.LocalUser:
                            # print(object_with_required_fields)
                            user_id = object_with_required_fields["unique_id"]
                            temp_oaa_obj = self._provider_object.local_users.get(
                                user_id
                            )

                            if not temp_oaa_obj:
                                temp_oaa_obj = self._provider_object.add_local_user(
                                    **object_with_required_fields
                                )  # noqa E501
                            # required_fields = oaa_utils.REQUIRED_FIELDS['LocalUser']
                        elif source.destination == VezaObjectType.User:
                            user_id = object_with_required_fields["identity"]
                            temp_oaa_obj = self._provider_object.users.get(user_id)

                            if not temp_oaa_obj:
                                # import bpdb; po = self._provider_object; bpdb.set_trace()  # noqa: E702
                                temp_oaa_obj = self._provider_object.add_user(
                                    **object_with_required_fields
                                )  # noqa E501
                        elif source.destination == VezaObjectType.Group:
                            temp_oaa_obj = self._provider_object.groups.get(
                                object_with_required_fields["unique_id"]
                            )

                            if not temp_oaa_obj:
                                temp_oaa_obj = self._provider_object.add_group(
                                    **object_with_required_fields
                                )  # noqa E501
                                # import bpdb; bpdb.set_trace()  # noqa: E702
                            # required_fields = oaa_utils.REQUIRED_FIELDS['LocalUser']

                        elif source.destination == VezaObjectType.LocalGroup:
                            temp_oaa_obj = self._provider_object.local_groups.get(
                                object_with_required_fields["unique_id"]
                            )

                            if not temp_oaa_obj:
                                temp_oaa_obj = self._provider_object.add_local_group(
                                    **object_with_required_fields
                                )  # noqa E501
                                # import bpdb; bpdb.set_trace()  # noqa: E702
                            # required_fields = oaa_utils.REQUIRED_FIELDS['LocalUser']
                        elif source.destination == VezaObjectType.Employee:
                            if (
                                object_with_required_fields["unique_id"]
                                not in self._provider_object.employees
                            ):
                                # print(object_with_required_fields)
                                # import bpdb; bpdb.set_trace()  # noqa: E702
                                temp_oaa_obj = self._provider_object.add_employee(
                                    **object_with_required_fields
                                )  # noqa E501

                        elif source.destination == VezaObjectType.LocalRole:
                            if (
                                object_with_required_fields["unique_id"]
                                not in self._provider_object.local_roles
                            ):
                                temp_oaa_obj = self._provider_object.add_local_role(
                                    **object_with_required_fields
                                )  # noqa E501
                        elif source.destination == VezaObjectType.Resource:
                            if (
                                object_with_required_fields["unique_id"]
                                not in self._provider_object.resources
                            ):
                                temp_oaa_obj = self._provider_object.add_resource(
                                    resource_type=source.resource_type,
                                    **object_with_required_fields,
                                )  # noqa E501
                        # Add other attributes

                        # These are those not already added as required
                        filtered_fields = filter(
                            lambda f: f not in required_fields_tmp, source.attributes
                        )  # noqa E501

                        if temp_oaa_obj:
                            for update_field in filtered_fields:
                                attr_source = source.attributes[update_field]

                                setattr(temp_oaa_obj, update_field, r[attr_source])

                            # custom properties are set using the `set_property` method
                            # with the name of the property and value

                            for property_name, f in source.custom_attrs_map.items():
                                source_field = f.source_field or f.destination_field

                                if r.get(source_field) in [None, ""]:
                                    logger.debug(
                                        "field [%s] is blank on %s", source_field, r.row
                                    )
                                else:
                                    logger.debug(
                                        "adding %s=%s to object [%s]",
                                        source_field,
                                        r[source_field],
                                        r.row,
                                    )

                                    temp_oaa_obj.set_property(
                                        property_name.lower(), r[source_field]
                                    )  # noqa E501

            logger.info(
                f"Finished loading {source.destination} for {source.source_name}"
            )



"""
Pre-flight validation functions for config.yaml validation.

This module provides standalone validation functions that can be used
to validate configuration files without requiring a full sync or Veza.
"""

import importlib
import logging

from pydantic import ValidationError

from oaa.settings import Config

logger = logging.getLogger(__name__)


def validate_config_structure(config_file):
    """Validate the config.yaml structure using Pydantic.

    :param config_file: Path to the config.yaml file
    :returns: Tuple of (is_valid, errors) where errors is a list of formatted error messages
    """
    errors = []

    try:
        Config.reset()
        Config.update(config_file=config_file)

        if Config._CONFIG:
            return True, []
        else:
            return False, ["Configuration failed to load"]

    except ValidationError as err:
        for e in err.errors():
            loc = (
                ".".join(str(loc_item) for loc_item in e["loc"]) if e["loc"] else "root"
            )  # noqa: E741
            msg = f"  - {loc}: {e['msg']}"
            errors.append(msg)
        return False, errors
    except Exception as err:
        return False, [str(err)]


def validate_connection_modules(config):
    """Validate that custom connection modules can be imported.

    :param config: Config object with connections
    :returns: Tuple of (is_valid, messages) where messages includes success/failure info
    """
    messages = []
    all_valid = True

    if not hasattr(config, "connections") or not config.connections:
        messages.append("No connections defined in config")
        return True, messages

    for name, conn in config.connections.items():
        if hasattr(conn, "custom_module") and conn.custom_module:
            module_import_paths = [
                conn.custom_module,
                f"oaa.modules.{conn.custom_module}",
            ]

            imported = False
            for module_path in module_import_paths:
                try:
                    importlib.import_module(module_path)
                    messages.append(
                        f'✅ Connection "{name}": module "{module_path}" imported successfully'
                    )
                    imported = True
                    break
                except ModuleNotFoundError:
                    continue
                except Exception as exc:
                    messages.append(
                        f'❌ Connection "{name}": error importing "{module_path}": {exc}'
                    )
                    all_valid = False
                    imported = True
                    break

            if not imported:
                messages.append(
                    f'❌ Connection "{name}": could not find module "{conn.custom_module}" '
                    f"(tried: {module_import_paths})"
                )
                all_valid = False

    return all_valid, messages


def test_connectivity(config, test_all=False):
    """Test connectivity to data sources defined in connections.

    :param config: Config object with connections
    :param test_all: If True, test all connections. If False, only test critical ones
    :returns: Tuple of (is_valid, messages) where messages includes success/failure info
    """
    messages = []
    all_valid = True

    if not hasattr(config, "connections") or not config.connections:
        messages.append("No connections defined in config")
        return True, messages

    for name, conn in config.connections.items():
        conn_type = getattr(conn, "connection_type", None)

        if conn_type == "db":
            messages.append(f'Testing DB connection "{name}"...')
            try:
                if hasattr(conn, "_get_connection"):
                    connection = conn._get_connection()
                    if connection:
                        messages.append(
                            f'✅ DB connection "{name}": connected successfully'
                        )
                        connection.close()
                    else:
                        messages.append(f'❌ DB connection "{name}": failed to connect')
                        all_valid = False
            except Exception as exc:
                messages.append(f'❌ DB connection "{name}": {exc}')
                all_valid = False

        elif conn_type in ("api", "veza_api"):
            messages.append(
                f'API connection "{name}": connectivity test not implemented '
                "(would require making actual API call)"
            )

        elif conn_type == "custom":
            messages.append(
                f'Custom connection "{name}": connectivity test not implemented '
                "(would require calling custom module fetch)"
            )

        elif conn_type == "csv":
            messages.append(f'CSV connection "{name}": checking file exists...')
            csv_path = getattr(conn, "csv_filepath", None)
            if csv_path:
                import os

                if os.path.exists(csv_path):
                    messages.append(f'✅ CSV file for connection "{name}": exists')
                else:
                    messages.append(
                        f'❌ CSV file for connection "{name}": file not found at {csv_path}'
                    )
                    all_valid = False

    return all_valid, messages


def run_validation_preflights(runner):
    """Run preflight checks suitable for validation mode.

    This runs the standard preflights but handles the case where
    config may not be fully loaded or Veza connection is not available.

    :param runner: CustomAppRunner instance
    :returns: Tuple of (is_valid, messages)
    """
    from oaa.hooks.preflight_checks import preflights

    messages = []
    all_valid = True

    for pf in preflights:
        pf_name = pf.__name__
        logger.debug(f"Running preflight: {pf_name}")

        try:
            result = pf(runner)

            if result:
                for issue in result:
                    messages.append(f"❌ {pf_name}: {issue}")
                    all_valid = False
            else:
                messages.append(f"✅ {pf_name}: passed")

        except Exception as exc:
            messages.append(f"⚠️  {pf_name}: error running check - {exc}")
            logger.debug(f"Preflight {pf_name} error: {exc}")

    return all_valid, messages


def validate_config(config_file, test_connectivity=False):
    """Main validation function that runs all validation steps.

    :param config_file: Path to the config.yaml file
    :param test_connectivity: Whether to test actual connectivity to data sources
    :returns: Tuple of (is_valid, all_messages) where all_messages is a combined list
    """
    all_messages = []
    all_valid = True

    all_messages.append("=" * 50)
    all_messages.append("STEP 1: Validating config structure...")
    all_messages.append("=" * 50)

    is_valid, errors = validate_config_structure(config_file)
    if not is_valid:
        all_valid = False
        all_messages.extend(errors)
        return all_valid, all_messages

    all_messages.append("✅ Config structure is valid")

    config = Config._CONFIG

    all_messages.append("")
    all_messages.append("=" * 50)
    all_messages.append("STEP 2: Validating connection modules...")
    all_messages.append("=" * 50)

    is_valid, messages = validate_connection_modules(config)
    if not is_valid:
        all_valid = False
    all_messages.extend(messages)

    all_messages.append("")
    all_messages.append("=" * 50)
    all_messages.append("STEP 3: Running preflight checks...")
    all_messages.append("=" * 50)

    from oaa.app_runners.custom_app import CustomAppRunner

    try:
        runner = CustomAppRunner(config_file_yaml=config_file)
        is_valid, messages = run_validation_preflights(runner)
        if not is_valid:
            all_valid = False
        all_messages.extend(messages)
    except Exception as exc:
        all_messages.append(f"⚠️  Could not run preflight checks: {exc}")

    if test_connectivity:
        all_messages.append("")
        all_messages.append("=" * 50)
        all_messages.append("STEP 4: Testing connectivity (per user request)...")
        all_messages.append("=" * 50)

        is_valid, messages = test_connectivity(config, test_all=True)
        if not is_valid:
            all_valid = False
        all_messages.extend(messages)

    all_messages.append("")
    all_messages.append("=" * 50)
    if all_valid:
        all_messages.append("✅ VALIDATION PASSED - Configuration appears ready")
    else:
        all_messages.append("❌ VALIDATION FAILED - Please review errors above")
    all_messages.append("=" * 50)

    return all_valid, all_messages



"""
OAA Modules - Reusable utilities for oaa-runner integrations.

Available modules:
    - BaseURLSession: requests.Session with automatic base URL joining
    - lazy_dict_table: Memory-efficient PETL table from dict generators
    - params_or_env: Get config values from env/connection/source params
    - ConfigurationError: Exception for configuration issues

Example:
    from oaa.modules import BaseURLSession, lazy_dict_table, params_or_env
    from oaa.modules.exceptions import ConfigurationError
"""

from .base_url_session import BaseURLSession
from .lazy_dicts import lazy_dict_table
from .params_or_env import params_or_env
from .exceptions import ConfigurationError

__all__ = [
    "BaseURLSession",
    "lazy_dict_table",
    "params_or_env",
    "ConfigurationError",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import fire
import logging

# from pathlib import Path

# utils.py contains field transform functions
# oaa_utils.py contains functions specifically related to oaa
# cli_runner.py is the CLI logic
# from . import cli_runner
# from oaa.app_runners.hris_single_source import HRISRunner
# from oaa.app_runners.idp_runner import IdPRunner
from oaa.app_runners.custom_app import CustomAppRunner
from . import oaa_utils

logger = logging.getLogger(__name__)

# Set the Provider Type to link HRIS employees to. This should be the type of
# the primary IdP for the environment
# Current Options include:
# ACTIVE_DIRECTORY
# ANY
# AZURE_AD
# CUSTOM
# GOOGLE_WORKSPACE
# OKTA
# ONE_LOGIN


# try:
#     IDP_PROVIDER_TYPE = getattr(IdPProviderType, IDP_PROVIDER_TYPE.upper())
# except AttributeError as err:

#     logger.error('Invalid IdPProviderType type %s:  '
#                  'Must be one of %s',
#                  IDP_PROVIDER_TYPE,
#                  oaa_utils.idp_provider_types())
#     logger.error(err)
#     exit(1)

# Provide a path to an icon.  Icon should either be SVG or PNG and less than
# 64KB
# NOTE This is not implemented in this script yet.
# HRIS_ICON_B64_PATH = ""

# GROUP_TYPE = os.getenv('GROUP_TYPE', 'Department')

# what is the unique identifier field?
# HRIS_UNIQUE_ID_FIELD = os.environ['HRIS_UNIQUE_ID_FIELD']  # , 'EMP_NO')

def main(cli_args=None):
    """TODO: Docstring for main.
    :returns: TODO

    """
    payload = {
            # 'hris': HRISRunner,
            # 'idp': IdPRunner,
            'custom_app': CustomAppRunner,
            'oaa_utils': oaa_utils.FUNCS,
        }

    if cli_args:
        fire.Fire(payload, cli_args)
    else:
        fire.Fire(payload)


def direct_run(cli_args=None):
    if cli_args:
        fire.Fire(CustomAppRunner)
    else:
        fire.Fire(CustomAppRunner, cli_args)


if __name__ == "__main__":
    main()



import os
import json
import sys
import uuid
import random
from datetime import datetime, timezone
import logging
from faker import Faker
from scim2_models import User, Group
logging.StreamHandler(sys.stdout)
logger = logging.getLogger(__name__)

faker = Faker()

def make_groups(count=10):

    for x in range(count):
        TMP_USERS_LIST = USERS[:]
        group_id = f'{uuid.uuid4()}'
        group_name = faker.catch_phrase()
        group_payload = {
          "id": group_id,
          "externalID": group_id,
          "meta":  {
            "created":"2013-04-16T09:10:45Z",
            # "modified":"2013-04-16T09:10:45Z"
          },
          "displayName": group_name,
          "members":  [ ],
          # "members":  {
          #   'totalResults': 100,
          #   'startSubIndex': 1,
          #   'subCount': 5,
          #   'Resources': [

          #   ],
          # },
          "schemas":  [
            "urn:scim:schemas:core:2.0:group"
          ]

        }

        '''
        for x in range(random.randint(0, len(USERS))):
            u = TMP_USERS_LIST.pop(random.choice(list(range(0,
                                                            len(TMP_USERS_LIST)))))
            user_obj = {
              "value": u['id'],
              "display": u['displayName']
            } 
            # group_payload['members']['Resources'].append(user_obj)
            group_payload['members'].append(user_obj)
        '''

        group = Group.model_validate(group_payload)
        # assert user.user_name == 'bjensen@example.com'
        # assert user.meta.created == datetime.datetime(
        #     2010, 1, 23, 4, 56, 22, tzinfo=datetime.timezone.utc
        # )
        yield group.model_dump()


def make_users(count=10):
    for x in range(count):
        logger.info(f'creating another user: {x}')
        profile = faker.profile()
        first_name = faker.first_name()
        last_name = faker.last_name() 
        profile['name'] = f'{first_name} {last_name}'
        domain = 'acme.com'
        username = f'{first_name}.{last_name}'
        # created = '2025-02-16 13:38:26'
        created = datetime.now(timezone.utc)
        user_id = f'{uuid.uuid4()}'
        user_payload = {
            'schemas': [
                'urn:ietf:params:scim:schemas:core:2.0:User',
                # 'urn:ietf:params:scim:schemas:core:2.0:Extended',
                ],
            'id': user_id,
            'userName': f'{username}@{domain}',
            'name': {
                'familyName': last_name,
                'givenName': first_name
            },
            # 'lastName': last_name,
            'display_name': f'{first_name} {last_name}',
            'active': True,
            'userName': username,
            "emails": [
                {"value": f"{username}@example.com", "type": "home"},
                {"value": f"{username}@example.com", "type": "work"},
                {"value": f"{username}@test.org", "type": "other"}
            ],
            # 'mailAlias': '',
            # 'mailServer': 'null',
            # 'firstName': first_name,
            # 'createdDate': created,
            # 'multiSession': False,
            # 'createdByUser': 'hrms',
            # 'modifiedByUser': 'admin',

            # 'schemas': [
            #     'urn:soffid:com.soffid.iam.api.User'
            # ],
            # 'modifiedDate': '2021-05-04 09:24:54',
            # 'attributes': {},
            'userType': 'I',
            # 'homeServer': 'null',
            # 'primaryGroupDescription': 'World Original',
            # 'primaryGroup': 'world',
            'meta': {
                'resourceType': 'User', 'created': created,
                'lastModified': '2011-05-13T04:42:34Z',
                'version': '3694e05e9dff590',
                'location': f'http://flask2:8080/v2/Users/{user_id}',
                # 'links': {
                #     'roleAccounts': f'http://flask2:8080/scim2/v2/appname/RoleAccount?user_id={user_id}&enabled=true',
                #     'groupUsers': f'http://flask2:8080/scim2/v2/appname/GroupUser?user_id={user_id}&enabled=true',
                #     'accounts': f'http://flask2:8080/scim2/v2/appname/Account?user_id={user_id}&enabled=true',
                # },
            },
        }

        user = User.model_validate(user_payload)
        # assert user.user_name == 'bjensen@example.com'
        # assert user.meta.created == datetime.datetime(
        #     2010, 1, 23, 4, 56, 22, tzinfo=datetime.timezone.utc
        # )
        yield user.model_dump()

users_filename = '/app/test-users2.db.json'
groups_filename = '/app/test-groups2.db.json'

if os.path.exists(users_filename):
    with open(users_filename) as _f:
        USERS = json.load(_f)
else:
    USERS = list(make_users(500))
    with open(users_filename, 'w+') as _f:
        json.dump(USERS, _f)

if os.path.exists(groups_filename):
    with open(groups_filename) as _f:
        GROUPS = json.load(_f)
else:
    GROUPS = list(make_groups(20))
    with open(groups_filename, 'w+') as _f:
        json.dump(GROUPS, _f)

logger.info('There are %s users in the DB', len(USERS))
# exit()
USERS_DB = {u['id']:u for u in USERS}

logger.info('There are %s groups in the DB', len(GROUPS))
# exit()
GROUPS_DB = {g['id']:g for g in GROUPS}



from oaa.models import FieldMapping
from oaa.models import DynamicTranformOptions_Enum
import pytest

YESES = [
        'y',
        'Y',
        'Yes',
        'True',
        '1',
        1
]


@pytest.fixture(params=YESES)
def include(request):
    yield request.param


@pytest.fixture(params=DynamicTranformOptions_Enum)
def transform_option(request):
    yield str(request.param.name)


@pytest.fixture(params=YESES)
def is_custom(request):
    yield request.param


@pytest.mark.parametrize('none_value', ['none', 'None', None])
def test_field_mapping_None(none_value, include, is_custom):
    mapping = {
        'include': include,
        'source_field': 'EMPL_NO',
        'veza_obj': 'Employee',
        'is_custom': is_custom,
        'destination_field': 'employee_number',
        'transform': none_value
    }

    mapping_obj = FieldMapping.model_validate(mapping)

    assert mapping_obj


def test_field_mapping_none(include, is_custom):
    mapping = {
        'include': include,
        'source_field': 'EMPL_NO',
        'veza_obj': 'Employee',
        'is_custom': is_custom,
        'destination_field': 'employee_number',
        'transform': 'None'
    }

    mapping_obj = FieldMapping.model_validate(mapping)

    assert mapping_obj


# def test_field_mapping_other_transform(include, is_custom, transform_option):
def test_field_mapping_other_transform(include, is_custom, transform_option):
    mapping = {
        'include': include,
        'source_field': 'EMPL_NO',
        'veza_obj': 'Employee',
        'is_custom': is_custom,
        'destination_field': 'employee_number',
        'transform': transform_option
    }

    mapping_obj = FieldMapping.model_validate(mapping)

    assert mapping_obj



import os
import pytest
from pathlib import Path
from oaa.app_runners.custom_app import CustomAppRunner
BASE_DIR = Path(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture
def _config(config):
    config.update(config_file=BASE_DIR / 'config-qvia-sample.yaml')

    return config


def test_source_should_have_multiple_transforms(_config):
    assert _config.config_file
    assert _config.sources
    source = _config.sources['users']
    assert len(source.field_mappings()['COMBINED_FOR_NAME'].transforms) == 1


def test_stream_and_get_stream_should_return_same_object(_config):

    source = _config.sources['users']

    assert id(source.stream) == id(source.get_stream())


def test_run_transforms_lowercase(_config):
    runner = CustomAppRunner()

    source = _config.sources['users']

    runner._run_transform(source)
    # runner.run()

    for r in source.records:
        assert 'EMAIL' in r.flds
        assert 'name' in r.flds
        assert r.name == r.name.lower()

        # import bpdb; bpdb.set_trace()  # noqa: E702
        # assert 'unique_id' in r.flds


def test_source2(_config):
    runner = CustomAppRunner()
    runner._run_transforms()
    source2 = _config.sources['application_mfa_type']
    print(source2._stream.fieldnames())

    for r in source2.records:
        assert 'mfa_type' in r.flds
        assert 'mfa_enabled' in r.flds
        assert 'mfa_enabled_for_external_and_fed' in r.flds
        assert 'mfa_enabled_for_internal' in r.flds
        assert 'unique_id' in r.flds


def test_application_as_resource(_config):
    runner = CustomAppRunner()
    runner.run(push_to_oaa=False)
    source2 = _config.sources['application_as_resource']
    print(source2._stream)

    for r in source2.records:
        assert 'unique_id' in r.flds
        assert 'name' in r.flds
        assert 'mfa_enabled_for_external_and_fed' in r.flds
        assert 'mfa_enabled_for_internal' in r.flds
        assert r['mfa_enabled_for_external_and_fed'] in (True, False)
        assert r['mfa_enabled_for_internal'] in (True, False)


def test_3_apps_resources(_config):
    runner = CustomAppRunner()

    runner.run(push_to_oaa=False)
    provider = runner._provider_object

    assert len(provider.resources) == 3


USERS = [
    ['iqviasda2@yahoo.com', 'EXT'],
    ['iqviasda3@federation.com', 'FED'],
    ['iqviasda4@iqvia.com', 'INT'],
]


@pytest.mark.parametrize('username,user_type', USERS)
def test_user_in_provider(username, user_type, _config):

    runner = CustomAppRunner()

    runner.run(push_to_oaa=False)

    provider = runner._provider_object

    assert username in provider.local_users
    user = provider.local_users[username]

    assert user.unique_id == username
    assert user.user_type == user_type


APPS = [
    ['Test OIDC Application 5', 'OIDC-OIDC-TST5', True, False],
    ['Test SAML Application', 'SAML-SAML-SSO-APP-SDA', True, False],
    ['Test OIDC Application', 'OIDC-OIDC-APP-SDA', False, True],
]


@pytest.mark.parametrize('application_name,application_id,mfa_enabled_for_external_and_fed,mfa_enabled_for_internal',
                         APPS)
def test_apps_in_resources(application_name, application_id,
                           mfa_enabled_for_external_and_fed,
                           mfa_enabled_for_internal, _config):

    runner = CustomAppRunner()

    runner.run(push_to_oaa=False)

    provider = runner._provider_object

    assert application_id in provider.resources


@pytest.mark.parametrize('application_name,application_id,mfa_enabled_for_external_and_fed,mfa_enabled_for_internal',
                         APPS)
def test_apps_in_resources(application_name, application_id,
                           mfa_enabled_for_external_and_fed,
                           mfa_enabled_for_internal, _config):
    runner = CustomAppRunner()

    runner.run(push_to_oaa=False)

    provider = runner._provider_object

    assert application_id in provider.resources
    app_resource = provider.resources[application_id]

    assert app_resource.unique_id == application_id
    assert app_resource.name == application_name
    resource_fields = provider.property_definitions.resource_properties['IQVIA Application'].keys()
    assert 'mfa_enabled_for_external_and_fed' in resource_fields
    assert 'mfa_enabled_for_internal' in resource_fields

    assert app_resource.properties
    # assert app_resource.properties['mfa_enabled_for_internal'] == mfa_enabled_for_internal
    # assert app_resource.properties['mfa_enabled_for_external_and_fed'] == mfa_enabled_for_external_and_fed
